import pickle
import six
import copy


def is_string(val):
    return isinstance(val, six.string_types)


def is_list(val):
    return type(val) in (list, tuple)


def is_dict(val):
    return isinstance(val, dict)


def make_list(val):
    return [val] if not is_list(val) else val


def make_string(val):
    return str(val) if not is_string(val) else val


def make_dict_from_list(val):
    return dict([(_, None) for _ in val])


def merge_dict(d1, d2):
    """
        d1 <-- d2
    """
    d = copy.deepcopy(d2)
    for k, v in list(d1.items()):
        if k not in d:
            d[k] = v
        else:
            if isinstance(v, dict):
                d[k] = merge_dict(d1[k], d[k])
    return d


def dict_path(from_d, to_d={}, l=[]):
    """
    Returns a dictionary with the path in which each of the keys is found

    Parameters:
        from_d : dict
            Dictionary that contains all the keys, values
        to_d : dict
            Dictionary to which the results will be appended

    Example:
        dict_path({'level1':{'level2':{'level3':'value'}}})
        Returns
            {'level1': [],
              'level2': ['level1'],
              'level3': ['level1', 'level2']
             }
    """
    for k, v in list(from_d.items()):
        if isinstance(v, dict):
            to_d[k] = l
            _l = copy.deepcopy(l)
            _l.append(k)
            to_d = dict_path(from_d[k], to_d, _l)
        else:
            to_d[k] = l
    _to_d = to_d.copy()
    to_d = {}
    return _to_d


def dict_update(d, k, val, d_ref=None):
    d_ref = d if not d_ref else d
    path = dict_path(d_ref)
    if path:
        reduce(dict.get, path[k], d).update({k: val})
    else:
        d.update(k=val)
    return d


def pp(el, preString=''):
    """
    Formats (prettyprint) a concatenated dictionary
    """
    tab = ' ' * 4
    if isinstance(el, dict):
        keys = list(el.keys())
        keys.sort()
        for key in keys:
            val = el[key]
            if isinstance(val, dict) or isinstance(val, list):
                print('%s%s :' % (preString, key))
                pp(val, preString + tab)
            else:
                print('%s%s  =  %s' % (preString, key, val))

    elif isinstance(el, list):
        print(preString + tab + '[')
        preString += tab
        for _ in el:
            if isinstance(_, dict):
                print(preString + tab + '{')
                pp(_, preString + tab * 2)
                print(preString + tab + '}')
            elif isinstance(_, list):
                print(preString + tab + '[')
                pp(_, preString + tab)
                print(preString + tab + ']')
            else:
                pp(_, preString + tab)
        # print preString+'      '+str('-')*10
        print(preString + ']')
    else:
        print(preString + str(el))


def inverseDict(d):
    """
    Returns a dictionay indexed by values {value_k:key_k}
    Parameters:
    -----------
        d : dictionary
    """
    dt = {}
    for k, v in list(d.items()):
        if type(v) in (list, tuple):
            for i in v:
                dt[i] = k
        else:
            dt[v] = k
    return dt


## Kwargs handlers

def kwargs_from_keyword(from_kwargs, to_kwargs, keyword, clean_origin=True):
    """
    Looks for keys of the format keyword_value.
    And return a dictionary with {keyword:value} format

    Parameters:
    -----------
        from_kwargs : dict
            Original dictionary
        to_kwargs : dict
            Dictionary where the items will be appended
        keyword : string
            Keyword to look for in the orginal dictionary
        clean_origin : bool
            If True then the k,v pairs from the original
            dictionary are deleted
    """
    for k in list(from_kwargs.keys()):
        if '{0}_'.format(keyword) in k:
            to_kwargs[k.replace('{0}_'.format(keyword), '')] = from_kwargs[k]
            if clean_origin:
                del from_kwargs[k]
    return to_kwargs


def check_kwargs(global_kwargs, values, local_kwargs=None, clean_origin=False):
    local_kwargs = {} if not local_kwargs else local_kwargs
    for kw in values:
        if kw in global_kwargs:
            local_kwargs[kw] = global_kwargs[kw]
            if clean_origin:
                del global_kwargs[kw]
    return local_kwargs


def dict_replace_keyword(d, keyword, kwargs, forceupdate=True):
    d = d.copy()
    for _k, _v in list(kwargs.items()):
        if keyword in _k:
            k = _k.replace(keyword, '')
            if keyword not in d:
                d[keyword] = {k: _v}
            else:
                if forceupdate:
                    d[keyword].update({k: _v})
                else:
                    if k not in d[keyword]:
                        d[keyword].update({k: _v})
        else:
            if forceupdate:
                d[_k] = _v
            else:
                if _k not in d:
                    d[_k] = _v

    return d


def save_pickle(obj, filename):
    """
    Serializes a given object
    Parameters:
    -----------
        obj : object
        filename : string
    """
    return pickle.dump(obj, open(filename, 'wb'))


def load_pickle(filename):
    """
    Loads a serialized object
    Parameters:
    -----------
        filename : string
    """
    return pickle.load(open(filename, 'rb'))


def deep_update(d, d_update):
    """
    Updates the values (deep form) of a given dictionary

    Parameters:
    -----------
        d : dict
            dictionary that contains the values to update
        d_update : dict
            dictionary to be updated
    """
    for k, v in list(d_update.items()):
        if isinstance(v, dict):
            if k in d:
                deep_update(d[k], v)
            else:
                d[k] = v
        elif isinstance(d, list):
            d.append({k: v})
        else:
            d[k] = v
    return d


def read_google(self, url, **kwargs):
    """
    Reads a google sheet
    """
    if url[-1] != '/':
        url += '/'
    return self.read_csv(url + 'export?gid=0&format=csv', **kwargs)



"""

QuantFigure allows you to create a persistent object. 
Annotations and Technical Studies can be added on demand. 

It accepts any dataframe with a timeseries index. 

Try it out:
	qf=cf.QuantFig(cf.datagen.ohlc())
	qf.iplot()

"""

from __future__ import absolute_import

import plotly.graph_objs as go
import json
import copy
import pandas as pd

from .plotlytools import iplot as pt_iplot
from . import tools
from . import ta
from . import utils
from . import colors
from . import auth
from . import date_tools

__QUANT_FIGURE_DATA = [
    'kind', 'showlegend', 'datalegend', 'name', 'slice', 'resample', 'bestfit',
    'text', 'title', 'yTitle', 'secondary_y_title', 'bestfit_colors', 'kind',
    'colorscale', 'xTitle', 'colors', 'secondary_y'
]

__QUANT_FIGURE_LAYOUT = [
    'annotations', 'showlegend', 'margin', 'rangeselector', 'rangeslider', 'shapes',
    'width', 'height', 'dimensions'
]

__QUANT_FIGURE_THEME = ['theme', 'up_color', 'down_color']
__QUANT_FIGURE_PANELS = ['min_panel_size', 'spacing', 'top_margin', 'bottom_margin']


def get_layout_kwargs():
    return tools.__LAYOUT_KWARGS


def get_annotation_kwargs():
    return tools.__ANN_KWARGS


def get_shapes_kwargs(): return tools.__SHAPES_KWARGS


class QuantFig(object):

    def __init__(self, df, kind='candlestick', columns=None, **kwargs):
        self.df = df
        self.studies = {}
        self.data = {}
        self.theme = {}
        self.panels = {}
        self.layout = {}
        self.trendlines = []
        self.kwargs = {}

        # Set column names
        if not columns:
            columns = {}
            for _ in ['open', 'high', 'low', 'close', 'volume']:
                columns[_] = kwargs.pop(_, '')
        self._d = ta._ohlc_dict(df, **columns)

        # Set initial annotations
        annotations = {
            'values': [],
            'params': utils.check_kwargs(kwargs, get_annotation_kwargs(), {}, clean_origin=True)
        }

        ann_values = kwargs.pop('annotations', None)

        if ann_values:
            if utils.is_list(ann_values):
                annotations['values'].extend(ann_values)
            else:
                annotations['values'].append(ann_values)

        # self.data initial values
        self.data.update(datalegend=kwargs.pop('datalegend', True), name=kwargs.pop('name', 'Trace 1'), kind=kind)
        self.data.update(slice=kwargs.pop('slice', (None, None)), resample=kwargs.pop('resample', None))

        # self.layout initial values
        self.layout['shapes'] = utils.check_kwargs(kwargs, get_shapes_kwargs(), {}, clean_origin=True)
        for k, v in list(self.layout['shapes'].items()):
            if not isinstance(v, list):
                self.layout['shapes'][k] = [v]
        self.layout['rangeselector'] = kwargs.pop('rangeselector', {'visible': False})
        self.layout['rangeslider'] = kwargs.pop('rangeslider', False)
        self.layout['margin'] = kwargs.pop('margin', dict(t=30, b=30, r=30, l=30))
        self.layout['annotations'] = annotations
        self.layout['showlegend'] = kwargs.pop('showlegend', True)
        self.layout.update(utils.check_kwargs(kwargs, get_layout_kwargs(), {}, clean_origin=True))

        # self.theme initial values
        self.theme['theme'] = kwargs.pop('theme', auth.get_config_file()['theme'])
        self.theme['up_color'] = kwargs.pop('up_color', '#17BECF')  # java
        self.theme['down_color'] = kwargs.pop('down_color', 'grey')

        # self.panels initial values
        self.panels['min_panel_size'] = kwargs.pop('min_panel_size', .15)
        self.panels['spacing'] = kwargs.pop('spacing', .08)
        self.panels['top_margin'] = kwargs.pop('top_margin', 0.9)
        self.panels['bottom_margin'] = kwargs.pop('top_margin', 0)
        self.update(**kwargs)

    def _get_schema(self):
        """
        Returns a dictionary with the schema for a QuantFigure

        """
        d = {}
        layout_kwargs = dict((_, '') for _ in get_layout_kwargs())
        for _ in ('data', 'layout', 'theme', 'panels'):
            d[_] = {}
            for __ in eval('__QUANT_FIGURE_{0}'.format(_.upper())):
                layout_kwargs.pop(__, None)
                d[_][__] = None
        d['layout'].update(annotations=dict(values=[],
                                            params=utils.make_dict_from_list(get_annotation_kwargs())))
        d['layout'].update(shapes=utils.make_dict_from_list(get_shapes_kwargs()))
        [layout_kwargs.pop(_, None) for _ in get_annotation_kwargs() + get_shapes_kwargs()]
        d['layout'].update(**layout_kwargs)
        return d

    def _get_sliced(self, slice, df=None, to_strfmt='%Y-%m-%d', from_strfmt='%d%b%y'):
        """
        Returns a sliced DataFrame

        Parameters
        ----------
            slice : tuple(from,to)
                from : str - ie ('01Jan2019')
                to : str - ie ('01Jan2020')
                    States the 'from' and 'to' values which
                    will get rendered as df.loc[from:to]
            df : DataFrame
                If omitted then the QuantFigure.DataFrame is resampled.
        """

        df = self.df.copy() if df == None else df
        if type(slice) not in (list, tuple):
            raise Exception('Slice must be a tuple two values')
        if len(slice) != 2:
            raise Exception('Slice must be a tuple two values')
        a, b = slice
        a = None if a in ('', None) else utils.make_string(a)
        b = None if b in ('', None) else utils.make_string(b)
        if a:
            a = date_tools.stringToString(a, from_strfmt, to_strfmt) if '-' not in a else a
        if b:
            b = date_tools.stringToString(b, from_strfmt, to_strfmt) if '-' not in b else b

        return df.loc[a:b]

    def _get_resampled(self, rule, how={'ohlc': 'last', 'volume': 'sum'}, df=None, **kwargs):
        """
        Returns a resampled DataFrame

        Parameters
        ----------
            rule : str
                the offset string or object representing target conversion
                for all aliases available see http://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases
            how : str or dict
                states the form in which the resampling will be done.
                Examples:
                    how={'volume':'sum'}
                    how='count'
            df : DataFrame
                If omitted then the QuantFigure.DataFrame is resampled.
            kwargs
                For more information see http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.resample.html

        """
        df = self.df.copy() if df is None else df
        if rule == None:
            return df
        else:
            if isinstance(how, dict):
                if 'ohlc' in how:
                    v = how.pop('ohlc')
                    for _ in ['open', 'high', 'low', 'close']:
                        how[_] = v
                _how = how.copy()
                for _ in _how:
                    if _ not in self._d:
                        del how[_]
            return df.resample(rule=rule, **kwargs).apply(how)

    def update(self, **kwargs):
        """
        Updates the values for a QuantFigure
        The key-values are automatically assigned to the correct
        section of the QuantFigure
        """
        if 'columns' in kwargs:
            self._d = ta._ohlc_dict(self.df, columns=kwargs.pop('columns', None))
        schema = self._get_schema()
        annotations = kwargs.pop('annotations', None)

        if annotations:
            self.layout['annotations']['values'] = utils.make_list(annotations)
        for k, v in list(kwargs.items()):
            try:
                utils.dict_update(self.__dict__, k, v, schema)
            except:
                self.kwargs.update({k: v})

    def delete(self, *args):
        """
        Deletes the values for a QuantFigure
        The key-values are automatically deleted from the correct
        section of the QuantFigure
        """
        if args:
            args = args[0] if utils.is_list(args[0]) else args
            path = utils.dict_path(self.__dict__)
            for _ in args:
                if _ in self.__dict__.keys():
                    raise Exception('"{0}" cannot be deleted'.format(_))

            for a in args:
                try:
                    if a in ('shapes'):
                        self.layout[a].clear()
                    elif a == 'annotations':
                        self.layout['annotations'] = {'values': [], 'params': {}}
                    else:
                        del reduce(dict.get, path[a], self.__dict__)[a]
                except:
                    raise Exception('Key: {0} not found'.format(a))

    def figure(self, **kwargs):
        """

        Returns a Plotly figure

        """
        kwargs['asFigure'] = True
        return self.iplot(**kwargs)

    def _panel_domains(self, n=2, min_panel_size=.15, spacing=0.08, top_margin=1, bottom_margin=0):
        """

        Returns the panel domains for each axis

        """
        d = {}
        for _ in range(n + 1, 1, -1):
            lower = round(bottom_margin + (min_panel_size + spacing) * (n + 1 - _), 2)
            d['yaxis{0}'.format(_)] = dict(domain=(lower, lower + min_panel_size))
        top = d['yaxis2']['domain']
        d['yaxis2']['domain'] = (top[0], top_margin)
        return d

    def _get_trendline(self, date0=None, date1=None, on=None, kind='trend', to_strfmt='%Y-%m-%d', from_strfmt='%d%b%y',
                       **kwargs):
        """
        Returns a trendline (line), support or resistance

        Parameters:
            date0 : string
                Trendline starting date
            date1 :  string
                Trendline end date
            on : string
                Indicate the data series in which the
                trendline should be based.
                    'close'
                    'high'
                    'low'
                    'open'
            kind : string
                Defines de kind of trendline
                    'trend'
                    'resistance'
                    'support'
            mode : string
                Defines how the support/resistance will
                be drawn
                    'starttoened' : (x0,x1)
                    'fromstart' : (x0,date0)
                    'toend'	: (date0,x1)
            text : string
                If passed, then an annotation will be added
                to the trendline (at mid point)
            from_strfmt : string
                Defines the date formating in which
                date0 and date1 are stated.
                default: '%d%b%y'
            to_strfmt : string
                Defines the date formatting
                to which it should be converted.
                This should match the same format as the timeseries index.
                default : '%Y-%m-%d'
        """
        ann_values = copy.deepcopy(get_annotation_kwargs())
        ann_values.extend(['x', 'y'])
        ann_kwargs = utils.check_kwargs(kwargs, ann_values, {}, clean_origin=True)

        def position(d0, d1):
            return d0 + (d1 - d0) / 2

        date0 = kwargs.pop('date', date0)
        date0 = date_tools.stringToString(date0, from_strfmt, to_strfmt) if '-' not in date0 else date0

        if kind == 'trend':
            date1 = date_tools.stringToString(date1, from_strfmt, to_strfmt) if '-' not in date1 else date1
            on = 'close' if not on else on
            df = pd.DataFrame(self.df[self._d[on]])
            y0 = kwargs.get('y0', df.loc[date0].values[0])
            y1 = kwargs.get('y1', df.loc[date1].values[0])

        if kind in ('support', 'resistance'):
            mode = kwargs.pop('mode', 'starttoend')
            if not on:
                on = 'low' if kind == 'support' else 'high'
            df = pd.DataFrame(self.df[self._d[on]])
            y0 = kwargs.get('y0', df.loc[date0].values[0])
            y1 = kwargs.get('y1', y0)
            if mode == 'starttoend':
                date0 = df.index[0]
                date1 = df.index[-1]
            elif mode == 'toend':
                date1 = df.index[-1]
            elif mode == 'fromstart':
                date1 = date0
                date0 = df.index[0]

        if isinstance(date0, pd.Timestamp):
            date0 = date_tools.dateToString(date0, to_strfmt)
        if isinstance(date1, pd.Timestamp):
            date1 = date_tools.dateToString(date1, to_strfmt)
        d = {'x0': date0, 'x1': date1, 'y0': y0, 'y1': y1}
        d.update(**kwargs)
        shape = tools.get_shape(**d)

        if ann_kwargs.get('text', False):
            ann_kwargs['x'] = ann_kwargs.get('x', date_tools.dateToString(
                position(date_tools.stringToDate(date0, to_strfmt), date_tools.stringToDate(date1, to_strfmt)),
                to_strfmt))
            ann_kwargs['y'] = ann_kwargs.get('y', position(shape['y0'], shape['y1']))
        else:
            ann_kwargs = {}
        return {'shape': shape, 'annotation': ann_kwargs}

    def add_trendline(self, date0, date1, on='close', text=None, **kwargs):
        """
        Adds a trendline to the QuantFigure.
        Given 2 dates, the trendline is connected on the data points
        that correspond to those dates.

        Parameters:
            date0 : string
                Trendline starting date
            date1 :  string
                Trendline end date
            on : string
                Indicate the data series in which the
                trendline should be based.
                    'close'
                    'high'
                    'low'
                    'open'
            text : string
                If passed, then an annotation will be added
                to the trendline (at mid point)

        kwargs:
            from_strfmt : string
                Defines the date formating in which
                date0 and date1 are stated.
                default: '%d%b%y'
            to_strfmt : string
                Defines the date formatting
                to which it should be converted.
                This should match the same format as the timeseries index.
                default : '%Y-%m-%d'
        """
        d = {'kind': 'trend', 'date0': date0, 'date1': date1, 'on': on, 'text': text}
        d.update(**kwargs)
        self.trendlines.append(d)

    def add_support(self, date, on='low', mode='starttoend', text=None, **kwargs):
        """
        Adds a support line to the QuantFigure

        Parameters:
            date0 : string
                The support line will be drawn at the 'y' level
                value that corresponds to this date.
            on : string
                Indicate the data series in which the
                support line should be based.
                    'close'
                    'high'
                    'low'
                    'open'
            mode : string
                Defines how the support/resistance will
                be drawn
                    'starttoened' : (x0,x1)
                    'fromstart' : (x0,date)
                    'toend' : (date,x1)
            text : string
                If passed, then an annotation will be added
                to the support line (at mid point)

        kwargs:
            from_strfmt : string
                Defines the date formating in which
                date0 and date1 are stated.
                default: '%d%b%y'
            to_strfmt : string
                Defines the date formatting
                to which it should be converted.
                This should match the same format as the timeseries index.
                default : '%Y-%m-%d'
        """
        d = {'kind': 'support', 'date': date, 'mode': mode, 'on': on, 'text': text}
        d.update(**kwargs)
        self.trendlines.append(d)

    def add_resistance(self, date, on='high', mode='starttoend', text=None, **kwargs):
        """
        Adds a resistance line to the QuantFigure

        Parameters:
            date0 : string
                The resistance line will be drawn at the 'y' level
                value that corresponds to this date.
            on : string
                Indicate the data series in which the
                resistance should be based.
                    'close'
                    'high'
                    'low'
                    'open'
            mode : string
                Defines how the support/resistance will
                be drawn
                    'starttoened' : (x0,x1)
                    'fromstart' : (x0,date)
                    'toend' : (date,x1)
            text : string
                If passed, then an annotation will be added
                to the resistance (at mid point)

        kwargs:
            from_strfmt : string
                Defines the date formating in which
                date0 and date1 are stated.
                default: '%d%b%y'
            to_strfmt : string
                Defines the date formatting
                to which it should be converted.
                This should match the same format as the timeseries index.
                default : '%Y-%m-%d'
        """
        d = {'kind': 'resistance', 'date': date, 'mode': mode, 'on': on, 'text': text}
        d.update(**kwargs)
        self.trendlines.append(d)

    def add_annotations(self, annotations, **kwargs):
        """
        Add an annotation to the QuantFigure.

        Parameters:
            annotations : dict or list(dict,)
                Annotations can be on the form form of
                    {'date' : 'text'}
                    and the text will automatically be placed at the
                    right level on the chart
                or
                    A Plotly fully defined annotation

        kwargs :
            fontcolor : str
                Text color for annotations
            fontsize : int
                Text size for annotations
            textangle : int
                Textt angle
            See https://plot.ly/python/reference/#layout-annotations
            for a complete list of valid parameters.

        """
        ann_kwargs = utils.check_kwargs(kwargs, get_annotation_kwargs(), {}, clean_origin=True)
        if type(annotations) == list:
            self.layout['annotations']['values'].extend(annotations)
        else:
            self.layout['annotations']['values'].append(annotations)
        if ann_kwargs:
            self.layout['annotations']['params'].update(**ann_kwargs)

    def add_shapes(self, **kwargs):
        """
        Add a shape to the QuantFigure.

        kwargs :
            hline : int, list or dict
                Draws a horizontal line at the
                indicated y position(s)
                Extra parameters can be passed in
                the form of a dictionary (see shapes)
            vline : int, list or dict
                Draws a vertical line at the
                indicated x position(s)
                Extra parameters can be passed in
                the form of a dictionary (see shapes)
            hspan : (y0,y1)
                Draws a horizontal rectangle at the
                indicated (y0,y1) positions.
                Extra parameters can be passed in
                the form of a dictionary (see shapes)
            vspan : (x0,x1)
                Draws a vertical rectangle at the
                indicated (x0,x1) positions.
                Extra parameters can be passed in
                the form of a dictionary (see shapes)
            shapes : dict or list(dict)
                List of dictionaries with the
                specifications of a given shape.
                See help(blot.tools.get_shape)
                for more information

        """
        kwargs = utils.check_kwargs(kwargs, get_shapes_kwargs(), {}, clean_origin=True)
        for k, v in list(kwargs.items()):
            if k in self.layout['shapes']:
                if utils.is_list(v):
                    self.layout['shapes'][k].extend(v)
                else:
                    self.layout['shapes'][k].append(v)
            else:
                self.layout['shapes'][k] = utils.make_list(v)

    # def add_study(self,name,params={}):
    # 	if 'kind' in params:
    # 			if params['kind'] in self._valid_studies:
    # 				self.studies[name]=params
    # 			else:
    # 				raise Exception('Invalid study: {0}'.format(params['kind']))
    # 	else:
    # 		raise Exception('Study kind required')

    def _add_study(self, study):
        """
        Adds a study to QuantFigure.studies

        Parameters:
            study : dict
                {'kind':study_kind,
                 'params':study_parameters,
                 'display':display_parameters}
        """
        str = '{study} {name}({period})' if study['params'].get('str', None) == None else study['params']['str']
        study['params']['str'] = str

        if not study['name']:
            study['name'] = ta.get_column_name(study['kind'].upper(), study=study['kind'],
                                               str=str,
                                               period=study['params'].get('periods', None),
                                               column=study['params'].get('column', None))

        restore = study['display'].pop('restore', False)

        if restore:
            _ = self.studies.pop(study['kind'], None)

        if study['kind'] in self.studies:
            try:
                id = '{0} ({1})'.format(study['kind'], study['params']['periods'])
            except:
                id = '{0} ({1})'.format(study['kind'], '(2)')
        else:
            id = study['kind']

        _id = id
        n = 1
        while id in self.studies:
            id = '{0} ({1})'.format(_id, n)
            n += 1
        self.studies[id] = study

    def add_volume(self, colorchange=True, column=None, name='', str='{name}', **kwargs):
        """
        Add 'volume' study to QuantFigure.studies

        Parameters:
            colorchange : bool
                If True then each volume bar will have a fill color
                depending on if 'base' had a positive or negative
                change compared to the previous value
                If False then each volume bar will have a fill color
                depending on if the volume data itself had a positive or negative
                change compared to the previous value
            column :string
                Defines the data column name that contains the volume data.
                Default: 'volume'
            name : string
                Name given to the study
            str : string
                Label factory for studies
                The following wildcards can be used:
                    {name} : Name of the column
                    {study} : Name of the study
                    {period} : Period used
                Examples:
                    'study: {study} - period: {period}'

        kwargs :
            base : string
                Defines the column which will define the
                positive/negative changes (if colorchange=True).
                Default = 'close'
            up_color : string
                Color for positive bars
            down_color : string
                Color for negative bars
        """
        if not column:
            column = self._d['volume']
        up_color = kwargs.pop('up_color', self.theme['up_color'])
        down_color = kwargs.pop('down_color', self.theme['down_color'])
        study = {'kind': 'volume',
                 'name': name,
                 'params': {'colorchange': colorchange, 'base': 'close', 'column': column,
                            'str': None},
                 'display': utils.merge_dict({'up_color': up_color, 'down_color': down_color}, kwargs)}
        self._add_study(study)

    def add_macd(self, fast_period=12, slow_period=26, signal_period=9, column=None,
                 name='', str=None, **kwargs):
        """
        Add Moving Average Convergence Divergence (MACD) study to QuantFigure.studies

        Parameters:
            fast_period : int
                MACD Fast Period
            slow_period : int
                MACD Slow Period
            signal_period : int
                MACD Signal Period
            column :string
                Defines the data column name that contains the
                data over which the study will be applied.
                Default: 'close'
            name : string
                Name given to the study
            str : string
                Label factory for studies
                The following wildcards can be used:
                    {name} : Name of the column
                    {study} : Name of the study
                    {period} : Period used
                Examples:
                    'study: {study} - period: {period}'
        kwargs:
            legendgroup : bool
                If true, all legend items are grouped into a
                single one
            All formatting values available on iplot()
        """

        if not column:
            column = self._d['close']
        study = {'kind': 'macd',
                 'name': name,
                 'params': {'fast_period': fast_period, 'slow_period': slow_period,
                            'signal_period': signal_period, 'column': column,
                            'str': str},
                 'display': utils.merge_dict({'legendgroup': False, 'colors': ['blue', 'red']}, kwargs)}
        study['params']['periods'] = '[{0},{1},{2}]'.format(fast_period, slow_period, signal_period)
        self._add_study(study)

    def add_sma(self, periods=20, column=None, name='',
                str=None, **kwargs):
        """
        Add Simple Moving Average (SMA) study to QuantFigure.studies

        Parameters:
            periods : int or list(int)
                Number of periods
            column :string
                Defines the data column name that contains the
                data over which the study will be applied.
                Default: 'close'
            name : string
                Name given to the study
            str : string
                Label factory for studies
                The following wildcards can be used:
                    {name} : Name of the column
                    {study} : Name of the study
                    {period} : Period used
                Examples:
                    'study: {study} - period: {period}'
        kwargs:
            legendgroup : bool
                If true, all legend items are grouped into a
                single one
            All formatting values available on iplot()
        """
        if not column:
            column = self._d['close']
        study = {'kind': 'sma',
                 'name': name,
                 'params': {'periods': periods, 'column': column,
                            'str': str},
                 'display': utils.merge_dict({'legendgroup': False}, kwargs)}
        self._add_study(study)

    def add_rsi(self, periods=20, rsi_upper=70, rsi_lower=30, showbands=True, column=None,
                name='', str=None, **kwargs):
        """
        Add Relative Strength Indicator (RSI) study to QuantFigure.studies

        Parameters:
            periods : int or list(int)
                Number of periods
            rsi_upper : int
                bounds [0,100]
                Upper (overbought) level
            rsi_lower : int
                bounds [0,100]
                Lower (oversold) level
            showbands : boolean
                If True, then the rsi_upper and
                rsi_lower levels are displayed
            column :string
                Defines the data column name that contains the
                data over which the study will be applied.
                Default: 'close'
            name : string
                Name given to the study
            str : string
                Label factory for studies
                The following wildcards can be used:
                    {name} : Name of the column
                    {study} : Name of the study
                    {period} : Period used
                Examples:
                    'study: {study} - period: {period}'
        kwargs:
            legendgroup : bool
                If true, all legend items are grouped into a
                single one
            All formatting values available on iplot()
        """
        if not column:
            column = self._d['close']
        str = str if str else '{name}({column},{period})'

        study = {'kind': 'rsi',
                 'name': name,
                 'params': {'periods': periods, 'column': column,
                            'str': str},
                 'display': utils.merge_dict({'legendgroup': True, 'rsi_upper': rsi_upper,
                                              'rsi_lower': rsi_lower, 'showbands': showbands}, kwargs)}
        self._add_study(study)

    def add_bollinger_bands(self, periods=20, boll_std=2, fill=True, column=None, name='',
                            str='{name}({column},{period})', **kwargs):
        """
        Add Bollinger Bands (BOLL) study to QuantFigure.studies

        Parameters:
            periods : int or list(int)
                Number of periods
            boll_std : int
                Number of standard deviations for
                the bollinger upper and lower bands
            fill : boolean
                If True, then the innner area of the
                bands will filled
            column :string
                Defines the data column name that contains the
                data over which the study will be applied.
                Default: 'close'
            name : string
                Name given to the study
            str : string
                Label factory for studies
                The following wildcards can be used:
                    {name} : Name of the column
                    {study} : Name of the study
                    {period} : Period used
                Examples:
                    'study: {study} - period: {period}'
        kwargs:
            legendgroup : bool
                If true, all legend items are grouped into a
                single one
            fillcolor : string
                Color to be used for the fill color.
                Example:
                    'rgba(62, 111, 176, .4)'
            All formatting values available on iplot()
        """
        if not column:
            column = self._d['close']
        study = {'kind': 'boll',
                 'name': name,
                 'params': {'periods': periods, 'boll_std': boll_std, 'column': column,
                            'str': str},
                 'display': utils.merge_dict({'legendgroup': True, 'fill': fill}, kwargs)}
        self._add_study(study)

    def add_ema(self, periods=20, column=None, str=None,
                name='', **kwargs):
        """
        Add Exponential Moving Average (EMA) study to QuantFigure.studies

        Parameters:
            periods : int or list(int)
                Number of periods
            column :string
                Defines the data column name that contains the
                data over which the study will be applied.
                Default: 'close'
            name : string
                Name given to the study
            str : string
                Label factory for studies
                The following wildcards can be used:
                    {name} : Name of the column
                    {study} : Name of the study
                    {period} : Period used
                Examples:
                    'study: {study} - period: {period}'
        kwargs:
            legendgroup : bool
                If true, all legend items are grouped into a
                single one
            All formatting values available on iplot()
        """
        if not column:
            column = self._d['close']
        study = {'kind': 'ema',
                 'name': name,
                 'params': {'periods': periods, 'column': column,
                            'str': str},
                 'display': utils.merge_dict({'legendgroup': False}, kwargs)}
        self._add_study(study)

    def add_cci(self, periods=14, cci_upper=100, cci_lower=-100,
                showbands=True, str=None, name='', **kwargs):
        """
        Commodity Channel Indicator study to QuantFigure.studies

        Parameters:
            periods : int or list(int)
                Number of periods
            cci_upper : int
                Upper bands level
                default : 100
            cci_lower : int
                Lower band level
                default : -100
            showbands : boolean
                If True, then the cci_upper and
                cci_lower levels are displayed
            name : string
                Name given to the study
            str : string
                Label factory for studies
                The following wildcards can be used:
                    {name} : Name of the column
                    {study} : Name of the study
                    {period} : Period used
                Examples:
                    'study: {study} - period: {period}'
        kwargs:
            legendgroup : bool
                If true, all legend items are grouped into a
                single one
            All formatting values available on iplot()
        """
        study = {'kind': 'cci',
                 'name': name,
                 'params': {'periods': periods, 'high': self._d['high'], 'low': self._d['low'],
                            'close': self._d['close'],
                            'str': str},
                 'display': utils.merge_dict({'legendgroup': True, 'cci_upper': cci_upper,
                                              'cci_lower': cci_lower, 'showbands': showbands}, kwargs)}
        self._add_study(study)

    def add_adx(self, periods=14, str=None, name='', **kwargs):
        """
        Add Average Directional Index (ADX) study to QuantFigure.studies

        Parameters:
            periods : int or list(int)
                Number of periods
            name : string
                Name given to the study
            str : string
                Label factory for studies
                The following wildcards can be used:
                    {name} : Name of the column
                    {study} : Name of the study
                    {period} : Period used
                Examples:
                    'study: {study} - period: {period}'
        kwargs:
            legendgroup : bool
                If true, all legend items are grouped into a
                single one
            All formatting values available on iplot()
        """
        study = {'kind': 'adx',
                 'name': name,
                 'params': {'periods': periods, 'high': self._d['high'], 'low': self._d['low'],
                            'close': self._d['close'],
                            'str': str},
                 'display': utils.merge_dict({'legendgroup': False}, kwargs)}
        self._add_study(study)

    def add_ptps(self, periods=14, af=0.2, initial='long', str=None, name='', **kwargs):
        """
        Add Parabolic SAR (PTPS) study to QuantFigure.studies

        Parameters:
            periods : int or list(int)
                Number of periods
            af : float
                acceleration factor
            initial : 'long' or 'short'
                Iniital position
                default: long
            name : string
                Name given to the study
            str : string
                Label factory for studies
                The following wildcards can be used:
                    {name} : Name of the column
                    {study} : Name of the study
                    {period} : Period used
                Examples:
                    'study: {study} - period: {period}'
        kwargs:
            legendgroup : bool
                If true, all legend items are grouped into a
                single one
            All formatting values available on iplot()
        """
        study = {'kind': 'ptps',
                 'name': name,
                 'params': {'periods': periods, 'high': self._d['high'], 'low': self._d['low'], 'af': af,
                            'initial': initial,
                            'str': str},
                 'display': utils.merge_dict({'legendgroup': False}, kwargs)}
        self._add_study(study)

    def add_atr(self, periods=14, str=None, name='', **kwargs):
        """
        Add Average True Range (ATR) study to QuantFigure.studies

        Parameters:
            periods : int or list(int)
                Number of periods
            name : string
                Name given to the study
            str : string
                Label factory for studies
                The following wildcards can be used:
                    {name} : Name of the column
                    {study} : Name of the study
                    {period} : Period used
                Examples:
                    'study: {study} - period: {period}'
        kwargs:
            legendgroup : bool
                If true, all legend items are grouped into a
                single one
            All formatting values available on iplot()
        """
        study = {'kind': 'atr',
                 'name': name,
                 'params': {'periods': periods, 'high': self._d['high'], 'low': self._d['low'],
                            'close': self._d['close'],
                            'str': str},
                 'display': utils.merge_dict({'legendgroup': False}, kwargs)}
        self._add_study(study)

    def add_dmi(self, periods=14, str='{name}({period})',
                name='', **kwargs):
        """
        Add Directional Movement Index (DMI) study to QuantFigure.studies

        Parameters:
            periods : int or list(int)
                Number of periods
            name : string
                Name given to the study
            str : string
                Label factory for studies
                The following wildcards can be used:
                    {name} : Name of the column
                    {study} : Name of the study
                    {period} : Period used
                Examples:
                    'study: {study} - period: {period}'
        kwargs:
            legendgroup : bool
                If true, all legend items are grouped into a
                single one
            All formatting values available on iplot()
        """
        study = {'kind': 'dmi',
                 'name': name,
                 'params': {'periods': periods, 'high': self._d['high'], 'low': self._d['low'],
                            'close': self._d['close'],
                            'str': str},
                 'display': utils.merge_dict({'legendgroup': False}, kwargs)}
        self._add_study(study)

    def _get_study_figure(self, study_id, **kwargs):
        study = copy.deepcopy(self.studies[study_id])
        kind = study['kind']
        display = study['display']
        display['theme'] = display.get('theme', self.theme['theme'])
        params = study['params']
        name = study['name']
        params.update(include=False)
        local_kwargs = {}
        _slice = kwargs.pop('slice', self.data.get('slice', (None, None)))
        _resample = kwargs.pop('resample', self.data.get('resample', None))

        df = self._get_sliced(_slice).copy()
        if _resample:
            if utils.is_list(_resample):
                df = self._get_resampled(*_resample, df=df)
            elif utils.is_dict(_resample):
                _resample.update(df=df)
                df = self._get_resampled(**_resample)
            else:
                df = self._get_resampled(_resample, df=df)

        def get_params(locals_list, params, display, append_study=True):
            locals_list.append('legendgroup')
            local_kwargs = utils.check_kwargs(display, locals_list, {}, True)
            display.update(kwargs)
            if append_study:
                display = dict([('study_' + k, v) for k, v in display.items()])
            params.update(display)
            return local_kwargs, params

        if kind == 'volume':
            bar_colors = []
            local_kwargs, params = get_params([], params, display, False)
            # Fix for 152
            base_column = params['base'] if params['colorchange'] else 'volume'
            base = df[self._d[base_column]]
            up_color = colors.normalize(display['up_color']) if 'rgba' not in display['up_color'] else display[
                'up_color']
            down_color = colors.normalize(display['down_color']) if 'rgba' not in display['down_color'] else display[
                'down_color']
            study_kwargs = utils.kwargs_from_keyword(kwargs, {}, 'study')

            for i in range(len(base)):
                if i != 0:
                    if base[i] > base[i - 1]:
                        bar_colors.append(up_color)
                    else:
                        bar_colors.append(down_color)
                else:
                    bar_colors.append(down_color)
            fig = df[params['column']].figure(kind='bar', theme=params['theme'], **kwargs)
            fig['data'][0].update(marker=dict(color=bar_colors, line=dict(color=bar_colors)),
                                  opacity=0.8)

        if kind in ('sma', 'ema', 'atr', 'adx', 'dmi', 'ptps'):
            local_kwargs, params = get_params([], params, display)
            fig = df.ta_figure(study=kind, **params)

        if kind == 'boll':
            local_kwargs, params = get_params(['fill', 'fillcolor'], params, display)
            fig = df.ta_figure(study=kind, **params)
            if local_kwargs['fill']:
                fillcolor = local_kwargs.pop('fillcolor', fig['data'][2]['line']['color'] or 'rgba(200,200,200,.1)')
                fillcolor = colors.to_rgba(fillcolor, .1)
                fig['data'][2].update(fill='tonexty', fillcolor=fillcolor)

        if kind == 'rsi':
            locals_list = ['rsi_lower', 'rsi_upper', 'showbands']
            local_kwargs, params = get_params(locals_list, params, display)
            fig = df.ta_figure(study=kind, **params)
        # del fig.layout['shapes']
        # if local_kwargs['showbands']:
        # 	up_color=kwargs.get('up_color',self.theme['up_color'])
        # 	down_color=kwargs.get('down_color',self.theme['down_color'])
        # 	for _ in ('rsi_lower','rsi_upper'):
        # 		trace=fig.data[0].copy()
        # 		trace.update(y=[local_kwargs[_] for x in trace['x']])
        # 		trace.update(name='')
        # 		color=down_color if 'lower' in _ else up_color
        # 		trace.update(line=dict(color=color,width=1))
        # 		fig.data.append(trace)

        if kind == 'cci':
            locals_list = ['cci_lower', 'cci_upper', 'showbands']
            local_kwargs, params = get_params(locals_list, params, display)
            fig = df.ta_figure(study=kind, **params)
        # del fig.layout['shapes']
        # if local_kwargs['showbands']:
        # 	up_color=kwargs.get('up_color',self.theme['up_color'])
        # 	down_color=kwargs.get('down_color',self.theme['down_color'])
        # 	for _ in ('cci_lower','cci_upper'):
        # 		trace=fig.data[0].copy()
        # 		trace.update(y=[local_kwargs[_] for x in trace['x']])
        # 		trace.update(name='')
        # 		color=down_color if 'lower' in _ else up_color
        # 		trace.update(line=dict(color=color,width=1))
        # 		fig.data.append(trace)

        if kind == 'macd':
            local_kwargs, params = get_params([], params, display)
            fig = df.ta_figure(study=kind, **params)

        if local_kwargs.get('legendgroup', False):
            for trace in fig['data']:
                trace['legendgroup'] = name
                trace['showlegend'] = False
            fig['data'][0].update(showlegend=True, name=name)

        ## Has Bands
        if kind in ('rsi', 'cci'):
            fig = tools.fig_to_dict(fig)
            _upper = '{0}_upper'.format(kind)
            _lower = '{0}_lower'.format(kind)
            del fig['layout']['shapes']
            if local_kwargs['showbands']:
                up_color = kwargs.get('up_color', self.theme['up_color'])
                down_color = kwargs.get('down_color', self.theme['down_color'])
                for _ in (_lower, _upper):
                    trace = copy.deepcopy(fig['data'][0])
                    trace.update(y=[local_kwargs[_] for x in trace['x']])
                    trace.update(name='')
                    color = down_color if 'lower' in _ else up_color
                    trace.update(line=dict(color=color, width=1))
                    fig['data'].append(trace)

        return fig

    def iplot(self, **kwargs):
        __QUANT_FIGURE_EXPORT = ['asFigure', 'asUrl', 'asImage', 'asPlot', 'display_image', 'validate',
                                 'sharing', 'online', 'filename', 'dimensions']

        layout = copy.deepcopy(self.layout)
        data = copy.deepcopy(self.data)
        self_kwargs = copy.deepcopy(self.kwargs)

        data['slice'] = kwargs.pop('slice', data.pop('slice', (None, None)))
        data['resample'] = kwargs.pop('resample', data.pop('resample', None))

        asFigure = kwargs.pop('asFigure', False)
        showstudies = kwargs.pop('showstudies', True)
        study_kwargs = utils.kwargs_from_keyword(kwargs, {}, 'study', True)
        datalegend = kwargs.pop('datalegend', data.pop('datalegend', data.pop('showlegend', True)))
        export_kwargs = utils.check_kwargs(kwargs, __QUANT_FIGURE_EXPORT)

        _slice = data.pop('slice')
        _resample = data.pop('resample')

        panel_data = {}
        for k in ['min_panel_size', 'spacing', 'top_margin', 'bottom_margin']:
            panel_data[k] = kwargs.pop(k, self.panels[k])

        d = self_kwargs
        df = self._get_sliced(_slice).copy()
        if _resample:
            if utils.is_list(_resample):
                df = self._get_resampled(*_resample, df=df)
            elif utils.is_dict(_resample):
                _resample.update(df=df)
                df = self._get_resampled(**_resample)
            else:
                df = self._get_resampled(_resample, df=df)

        annotations = layout.pop('annotations')
        shapes = layout.pop('shapes')
        if not 'shapes' in shapes:
            shapes['shapes'] = []
        for trend in self.trendlines:
            _trend = self._get_trendline(**trend)
            shapes['shapes'].append(_trend['shape'])
            if 'text' in _trend['annotation']:
                annotations['values'].append(_trend['annotation'])
        shape_kwargs = utils.check_kwargs(kwargs, get_shapes_kwargs(), {}, clean_origin=True)
        for k, v in list(shape_kwargs.items()):
            if k in shapes:
                if isinstance(v, list):
                    shapes[k].extend(v)
                else:
                    shapes[k].append(v)
            else:
                shapes[k] = [v]
        for _ in [data, layout, self._d,
                  self.theme, {'annotations': annotations['values']},
                  annotations['params'], shapes]:
            if _:
                d = utils.merge_dict(d, _)
        d = utils.deep_update(d, kwargs)
        d = tools.updateColors(d)
        fig = df.figure(**d)

        if d['kind'] not in ('candle', 'candlestick', 'ohlc'):
            tools._move_axis(fig, yaxis='y2')  # FIXME TKP
            pass
        else:
            if not datalegend:
                fig['data'][0]['decreasing'].update(showlegend=False)
                fig['data'][0]['increasing'].update(showlegend=False)

        ## 126 Shapes in wrong axis
        for shape in fig['layout']['shapes']:
            if 'yref' in shape:
                if len(shape['yref']) == 1:  # not an explicity yref
                    shape.update(yref='y2')

        panel_data['n'] = 1

        which = [x['yaxis'] for x in fig['data']]
        which.sort()
        max_panel = int(which[-1][1:])
        figures = []

        if showstudies:
            kwargs = utils.check_kwargs(kwargs, ['theme', 'up_color', 'down_color'], {}, False)
            kwargs.update(**study_kwargs)
            kwargs.update(slice=_slice, resample=_resample)
            for k, v in list(self.studies.items()):
                study_fig = self._get_study_figure(k, **kwargs)
                study_fig = tools.fig_to_dict(study_fig)
                if 'yaxis' in study_fig['layout']:
                    study_fig['layout']['yaxis1'] = study_fig['layout']['yaxis'].copy()
                    del study_fig['layout']['yaxis']
                if v['kind'] in ('boll', 'sma', 'ema', 'ptps'):
                    tools._move_axis(study_fig, yaxis='y2')  # FIXME TKP
                    pass
                if v['kind'] in ('rsi', 'volume', 'macd', 'atr', 'adx', 'cci', 'dmi'):
                    max_panel += 1
                    panel_data['n'] += 1
                    tools._move_axis(study_fig, yaxis='y{0}'.format(max_panel))  # FIXME TKP
                figures.append(study_fig)
            figures.append(fig)
            fig = tools.merge_figures(figures)

            try:
                fig['layout']['xaxis1']['anchor'] = 'y2'
            except:
                fig['layout']['xaxis']['anchor'] = 'y2'

        domains = self._panel_domains(**panel_data)
        try:
            for k, v in list(domains.items()):
                fig['layout'][k].update(v)
        except:
            fig['layout'].update(**domains)
        if not d.get('rangeslider', False):
            try:
                del fig['layout']['yaxis1']
            except:
                pass
        if asFigure:
            return go.Figure(fig)
        else:
            return pt_iplot(fig, **export_kwargs)

    def __getitem__(self, key):
        return self.__dict__[key]

    def __repr__(self):
        _d = self.__dict__.copy()
        del _d['df']
        return json.dumps(_d, sort_keys=True, indent=4)



import pandas as pd


def to_df(figure):
    """
    Extracts the data from a Plotly Figure

    Parameters
    ----------
        figure : plotly_figure
            Figure from which data will be 
            extracted
        
        Returns a DataFrame or list of DataFrame
    """
    dfs = []
    for trace in figure['data']:
        if 'scatter' in trace['type']:
            try:
                if type(trace['x'][0]) == float:
                    index = trace['x']
                else:
                    index = pd.to_datetime(trace['x'])
            except:
                index = trace['x']
            if 'marker' in trace:
                d = {}
                if 'size' in trace['marker']:
                    size = trace['marker']['size']
                    if type(size) != list:
                        size = [size] * len(index)
                    d['size'] = size
                if 'text' in trace:
                    d['text'] = trace['text']
                if 'name' in trace:
                    name = trace['name']
                    if type(name) != list:
                        name = [name] * len(index)
                    d['categories'] = name
                d['y'] = trace['y']
                d['x'] = trace['x']
                if 'z' in trace:
                    d['z'] = trace['z']
                df = pd.DataFrame(d)
            else:
                df = pd.Series(trace['y'], index=index, name=trace['name'])
            dfs.append(df)
        elif trace['type'] in ('heatmap', 'surface'):
            df = pd.DataFrame(trace['z'].transpose(), index=trace['x'], columns=trace['y'])
            dfs.append(df)
        elif trace['type'] in ('box', 'histogram'):
            vals = trace['x'] if 'x' in trace else trace['y']
            df = pd.DataFrame({trace['name']: vals})
            dfs.append(df)
    if max(list(map(len, dfs))) == min(list(map(len, dfs))):
        if len(dfs) == 1:
            return dfs[0]
        else:
            if type(dfs[0]) == pd.core.series.Series:
                return pd.concat(dfs, axis=1)
            if all(dfs[0].columns == dfs[1].columns):
                return pd.concat(dfs, axis=0)
            else:
                return pd.concat(dfs, axis=1)
    else:
        try:
            return pd.concat(dfs)
        except:
            return dfs



##
# Special thanks to @krey for the python3 support
##

import numpy as np
import colorsys
import colorlover as cl
import operator
import copy

from collections import deque
from six import string_types
from IPython.display import HTML, display

from .utils import inverseDict
from .auth import get_config_file
from .exceptions import BlotError




def to_rgba(color, alpha):
    """
    Converts from hex|rgb to rgba

    Parameters:
    -----------
            color : string
                    Color representation on hex or rgb
            alpha : float
                    Value from 0 to 1.0 that represents the
                    alpha value.

    Example:
            to_rgba('#E1E5ED',0.6)
            to_rgba('#f03',0.7)
            to_rgba('rgb(23,23,23)',.5)
    """
    if type(color) == tuple:
        color, alpha = color
    color = color.lower()
    if 'rgba' in color:
        cl = list(eval(color.replace('rgba', '')))
        if alpha:
            cl[3] = alpha
        return 'rgba' + str(tuple(cl))
    elif 'rgb' in color:
        r, g, b = eval(color.replace('rgb', ''))
        return 'rgba' + str((r, g, b, alpha))
    else:
        return to_rgba(hex_to_rgb(color), alpha)


def hex_to_rgb(color):
    """
    Converts from hex to rgb

    Parameters:
    -----------
            color : string
                    Color representation on hex or rgb

    Example:
            hex_to_rgb('#E1E5ED')
            hex_to_rgb('#f03')
    """
    color = normalize(color)
    color = color[1:]
    # return 'rgb'+str(tuple(ord(c) for c in color.decode('hex')))
    return 'rgb' + str((int(color[0:2], base=16), int(color[2:4], base=16), int(color[4:6], base=16)))


def normalize(color):
    """
    Returns an hex color

    Parameters:
    -----------
            color : string
                    Color representation in rgba|rgb|hex

    Example:
            normalize('#f03')
    """
    if type(color) == tuple:
        color = to_rgba(*color)
    if 'rgba' in color:
        return rgb_to_hex(rgba_to_rgb(color))
    elif 'rgb' in color:
        return rgb_to_hex(color)
    elif '#' in color:
        if len(color) == 7:
            return color
        else:
            color = color[1:]
            return '#' + ''.join([x * 2 for x in list(color)])
    else:
        try:
            return normalize(cnames[color.lower()])
        except:
            raise BlotError('Not a valid color: ' + color)


def rgb_to_hex(color):
    """
    Converts from rgb to hex

    Parameters:
    -----------
            color : string
                    Color representation on hex or rgb

    Example:
            rgb_to_hex('rgb(23,25,24)')
    """
    rgb = eval(color.replace('rgb', ''))
    # return '#'+''.join(map(chr, rgb)).encode('hex')
    return '#' + ''.join(['{0:02x}'.format(x).upper() for x in rgb])


def rgba_to_rgb(color, bg='rgb(255,255,255)'):
    """
    Converts from rgba to rgb

    Parameters:
    -----------
            color : string
                    Color representation in rgba
            bg : string
                    Color representation in rgb

    Example:
            rgba_to_rgb('rgb(23,25,24,.4)''
    """

    def c_tup(c):
        return eval(c[c.find('('):])

    color = c_tup(color)
    bg = hex_to_rgb(normalize(bg))
    bg = c_tup(bg)
    a = color[3]
    r = [int((1 - a) * bg[i] + a * color[i]) for i in range(3)]
    return 'rgb' + str(tuple(r))


def hex_to_hsv(color):
    """
    Converts from hex to hsv

    Parameters:
    -----------
            color : string
                    Color representation on color

    Example:
            hex_to_hsv('#ff9933')
    """
    color = normalize(color)
    color = color[1:]
    # color=tuple(ord(c)/255.0 for c in color.decode('hex'))
    color = (int(color[0:2], base=16) / 255.0, int(color[2:4],
                                                   base=16) / 255.0, int(color[4:6], base=16) / 255.0)
    return colorsys.rgb_to_hsv(*color)


def color_range(color, N=20):
    """
    Generates a scale of colours from a base colour

    Parameters:
    -----------
            color : string
                    Color representation in hex
            N   : int
                    number of colours to generate

    Example:
            color_range('#ff9933',20)
    """
    color = normalize(color)
    org = color
    color = hex_to_hsv(color)
    HSV_tuples = [(color[0], x, color[2]) for x in np.arange(0, 1, 2.0 / N)]
    HSV_tuples.extend([(color[0], color[1], x)
                       for x in np.arange(0, 1, 2.0 / N)])
    hex_out = []
    for c in HSV_tuples:
        c = colorsys.hsv_to_rgb(*c)
        c = [int(_ * 255) for _ in c]
        # hex_out.append("#"+"".join([chr(x).encode('hex') for x in c]))
        hex_out.append("#" + "".join(['{0:02x}'.format(x) for x in c]))
    if org not in hex_out:
        hex_out.append(org)
    hex_out.sort()
    return hex_out


def color_table(color, N=1, sort=False, sort_values=False, inline=False, as_html=False):
    """
    Generates a colour table

    Parameters:
    -----------
            color : string | list | dict
                    Color representation in rgba|rgb|hex
                    If a list of colors is passed then these
                    are displayed in a table
            N   : int
                    number of colours to generate
                    When color is not a list then it generaes
                    a range of N colors
            sort : bool
                    if True then items are sorted
            sort_values : bool
                    if True then items are sorted by color values.
                    Only applies if color is a dictionary
            inline : bool
                    if True it returns single line color blocks
            as_html : bool
                    if True it returns the HTML code

    Example:
            color_table('#ff9933')
            color_table(blot.cnames)
            color_table(['pink','salmon','yellow'])
    Note:
            This function only works in iPython Notebook
    """
    if isinstance(color, list):
        c_ = ''
        rgb_tup = [normalize(c) for c in color]
        if sort:
            rgb_tup.sort()
    elif isinstance(color, dict):
        c_ = ''
        items = [(k, normalize(v), hex_to_hsv(normalize(v)))
                 for k, v in list(color.items())]
        if sort_values:
            items = sorted(items, key=operator.itemgetter(2))
        elif sort:
            items = sorted(items, key=operator.itemgetter(0))
        rgb_tup = [(k, v) for k, v, _ in items]
    else:
        c_ = normalize(color)
        if N > 1:
            rgb_tup = np.array(color_range(c_, N))[::-1]
        else:
            rgb_tup = [c_]

    def _color(c):
        if hex_to_hsv(c)[2] < .5:
            color = "#ffffff"
            shadow = '0 1px 0 #000'
        else:
            color = "#000000"
            shadow = '0 1px 0 rgba(255,255,255,0.6)'
        if c == c_:
            border = " border: 1px solid #ffffff;"
        else:
            border = ''
        return color, shadow, border

    s = '<ul style="list-style-type: none;">' if not inline else ''
    for c in rgb_tup:
        if isinstance(c, tuple):
            k, c = c
            k += ' : '
        else:
            k = ''
        if inline:
            s += '<div style="background-color:{0};height:20px;width:20px;display:inline-block;"></div>'.format(
                c)
        else:
            color, shadow, border = _color(c)
            s += """<li style="text-align:center;""" + border + """line-height:30px;background-color:""" + c + """;">
            <span style=" text-shadow:""" + shadow + """; color:""" + color + """;">""" + k + c.upper() + """</span>
            </li>"""
    s += '</ul>' if not inline else ''
    if as_html:
        return s
    return display(HTML(s))


def colorgen(colors=None, n=None, scale=None, theme=None):
    """
    Returns a generator with a list of colors
    and gradients of those colors

    Parameters:
    -----------
            colors : list(colors)
                    List of colors to use

    Example:
            colorgen()
            colorgen(['blue','red','pink'])
            colorgen(['#f03','rgb(23,25,25)'])
    """
    from .themes import THEMES
    step = .1
    if not colors:
        if not scale:
            if not theme:
                scale = get_config_file()['colorscale']
            else:
                scale = THEMES[theme]['colorscale']
        colors = get_scales(scale)
    dq = deque(colors)
    if len(dq) == 0:
        dq = deque(get_scales('ggplot'))
    if n:
        step = len(dq) * 0.8 / n if len(dq) * 8 < n else .1
    for i in np.arange(.2, 1, step):
        for y in dq:
            yield to_rgba(y, 1 - i + .2)
        dq.rotate(1)


# NEW STUFF


# Color Names
# ---------------------------------

cnames = {'aliceblue': '#F0F8FF',
          'antiquewhite': '#FAEBD7',
          'aqua': '#00FFFF',
          'aquamarine': '#7FFFD4',
          'azure': '#F0FFFF',
          'beige': '#F5F5DC',
          'bisque': '#FFE4C4',
          'black': '#000000',
          'blanchedalmond': '#FFEBCD',
          'blue': '#3780bf',
          'bluegray': '#565656',
          'bluepurple': '#6432AB',
          'blueviolet': '#8A2BE2',
          'brick': '#E24A33',
          'brightblue': '#0000FF',
          'brightred': '#FF0000',
          'brown': '#A52A2A',
          'burlywood': '#DEB887',
          'cadetblue': '#5F9EA0',
          'charcoal': '#151516',
          'chartreuse': '#7FFF00',
          'chocolate': '#D2691E',
          'coral': '#FF7F50',
          'cornflowerblue': '#6495ED',
          'cornsilk': '#FFF8DC',
          'crimson': '#DC143C',
          'cyan': '#00FFFF',
          'darkblue': '#00008B',
          'darkcyan': '#008B8B',
          'darkgoldenrod': '#B8860B',
          'darkgray': '#A9A9A9',
          'darkgreen': '#006400',
          'darkgrey': '#A9A9A9',
          'darkkhaki': '#BDB76B',
          'darkmagenta': '#8B008B',
          'darkolivegreen': '#556B2F',
          'darkorange': '#FF8C00',
          'darkorchid': '#9932CC',
          'darkred': '#8B0000',
          'darksalmon': '#E9967A',
          'darkseagreen': '#8FBC8F',
          'darkslateblue': '#483D8B',
          'darkslategray': '#2F4F4F',
          'darkslategrey': '#2F4F4F',
          'darkturquoise': '#00CED1',
          'darkviolet': '#9400D3',
          'deeppink': '#FF1493',
          'deepskyblue': '#00BFFF',
          'dimgray': '#696969',
          'dimgrey': '#696969',
          'dodgerblue': '#1E90FF',
          'firebrick': '#B22222',
          'floralwhite': '#FFFAF0',
          'forestgreen': '#228B22',
          'fuchsia': '#FF00FF',
          'gainsboro': '#DCDCDC',
          'ghostwhite': '#F8F8FF',
          'gold': '#FFD700',
          'goldenrod': '#DAA520',
          'grassgreen': '#32ab60',
          'gray': '#808080',
          'green': '#008000',
          'greenyellow': '#ADFF2F',
          'grey': '#808080',
          'grey01': '#0A0A0A',
          'grey02': '#151516',
          'grey03': '#1A1A1C',
          'grey04': '#1E1E21',
          'grey05': '#252529',
          'grey06': '#36363C',
          'grey07': '#3C3C42',
          'grey08': '#434343',
          'grey09': '#666570',
          'grey10': '#666666',
          'grey11': '#8C8C8C',
          'grey12': '#C2C2C2',
          'grey13': '#E2E2E2',
          'grey14': '#E5E5E5',
          'honeydew': '#F0FFF0',
          'hotpink': '#FF69B4',
          'indianred': '#CD5C5C',
          'indigo': '#4B0082',
          'ivory': '#FFFFF0',
          'java': '#17BECF',
          'khaki': '#F0E68C',
          'lavender': '#E6E6FA',
          'lavenderblush': '#FFF0F5',
          'lawngreen': '#7CFC00',
          'lemonchiffon': '#FFFACD',
          'lightpink2': '#fccde5',
          'lightpurple': '#bc80bd',
          'lightblue': '#ADD8E6',
          'lightcoral': '#F08080',
          'lightcyan': '#E0FFFF',
          'lightgoldenrodyellow': '#FAFAD2',
          'lightgray': '#D3D3D3',
          'lightgreen': '#90EE90',
          'lightgrey': '#D3D3D3',
          'lightivory': '#F6F6F6',
          'lightpink': '#FFB6C1',
          'lightsalmon': '#FFA07A',
          'lightseagreen': '#20B2AA',
          'lightskyblue': '#87CEFA',
          'lightslategray': '#778899',
          'lightslategrey': '#778899',
          'lightsteelblue': '#B0C4DE',
          'lightteal': '#8dd3c7',
          'lightyellow': '#FFFFE0',
          'lightblue2': '#80b1d3',
          'lightviolet': '#8476CA',
          'lime': '#00FF00',
          'lime2': '#8EBA42',
          'limegreen': '#32CD32',
          'linen': '#FAF0E6',
          'magenta': '#FF00FF',
          'maroon': '#800000',
          'mediumaquamarine': '#66CDAA',
          'mediumblue': '#0000CD',
          'mediumgray': '#656565',
          'mediumorchid': '#BA55D3',
          'mediumpurple': '#9370DB',
          'mediumseagreen': '#3CB371',
          'mediumslateblue': '#7B68EE',
          'mediumspringgreen': '#00FA9A',
          'mediumturquoise': '#48D1CC',
          'mediumvioletred': '#C71585',
          'midnightblue': '#191970',
          'mintcream': '#F5FFFA',
          'mistyrose': '#FFE4E1',
          'moccasin': '#FFE4B5',
          'mustard': '#FBC15E',
          'navajowhite': '#FFDEAD',
          'navy': '#000080',
          'oldlace': '#FDF5E6',
          'olive': '#808000',
          'olivedrab': '#6B8E23',
          'orange': '#ff9933',
          'orangered': '#FF4500',
          'orchid': '#DA70D6',
          'palegoldenrod': '#EEE8AA',
          'palegreen': '#98FB98',
          'paleolive': '#b3de69',
          'paleturquoise': '#AFEEEE',
          'palevioletred': '#DB7093',
          'papayawhip': '#FFEFD5',
          'peachpuff': '#FFDAB9',
          'pearl': '#D9D9D9',
          'pearl02': '#F5F6F9',
          'pearl03': '#E1E5ED',
          'pearl04': '#9499A3',
          'pearl05': '#6F7B8B',
          'pearl06': '#4D5663',
          'peru': '#CD853F',
          'pink': '#ff0088',
          'pinksalmon': '#FFB5B8',
          'plum': '#DDA0DD',
          'polar': '#ACAFB5',
          'polarblue': '#0080F0',
          'polarbluelight': '#46A0F0',
          'polarcyan': '#ADFCFC',
          'polardark': '#484848',
          'polardiv': '#D5D8DB',
          'polardust': '#F2F3F7',
          'polargrey': '#505050',
          'polargreen': '#309054',
          'polarorange': '#EE7600',
          'polarpurple': '#6262DE',
          'polarred': '#D94255',
          'powderblue': '#B0E0E6',
          'purple': '#800080',
          'red': '#db4052',
          'rose': '#FFC0CB',
          'rosybrown': '#BC8F8F',
          'royalblue': '#4169E1',
          'saddlebrown': '#8B4513',
          'salmon': '#fb8072',
          'sandybrown': '#FAA460',
          'seaborn': '#EAE7E4',
          'seagreen': '#2E8B57',
          'seashell': '#FFF5EE',
          'sienna': '#A0522D',
          'silver': '#C0C0C0',
          'skyblue': '#87CEEB',
          'slateblue': '#6A5ACD',
          'slategray': '#708090',
          'slategrey': '#708090',
          'smurf': '#3E6FB0',
          'snow': '#FFFAFA',
          'springgreen': '#00FF7F',
          'steelblue': '#4682B4',
          'tan': '#D2B48C',
          'teal': '#008080',
          'thistle': '#D8BFD8',
          'tomato': '#FF6347',
          'turquoise': '#40E0D0',
          'violet': '#EE82EE',
          'wheat': '#F5DEB3',
          'white': '#FFFFFF',
          'whitesmoke': '#F5F5F5',
          'yellow': '#ffff33',
          'yellowgreen': '#9ACD32',
          "henanigans_bg": "#242424",
          "henanigans_blue1": "#5F95DE",
          "henanigans_blue2": "#93B6E6",
          "henanigans_cyan1": "#7EC4CF",
          "henanigans_cyan2": "#B6ECF3",
          "henanigans_dark1": "#040404",
          "henanigans_dark2": "#141414",
          "henanigans_dialog1": "#444459",
          "henanigans_dialog2": "#5D5D7A",
          "henanigans_green1": "#8BD155",
          "henanigans_green2": "#A0D17B",
          "henanigans_grey1": "#343434",
          "henanigans_grey2": "#444444",
          "henanigans_light1": "#A4A4A4",
          "henanigans_light2": "#F4F4F4",
          "henanigans_orange1": "#EB9E58",
          "henanigans_orange2": "#EBB483",
          "henanigans_purple1": "#C98FDE",
          "henanigans_purple2": "#AC92DE",
          "henanigans_red1": "#F77E70",
          "henanigans_red2": "#DE958E",
          "henanigans_yellow1": "#E8EA7E",
          "henanigans_yellow2": "#E9EABE"
          }

# Custom Color Scales
# ---------------------------------

_custom_scales = {
    'qual': {
        # dflt only exists to keep backward compatibility after issue 91
        'dflt': ['orange', 'blue', 'grassgreen', 'purple', 'red', 'teal', 'yellow', 'olive', 'salmon', 'lightblue2'],
        'original': ['orange', 'blue', 'grassgreen', 'purple', 'red', 'teal', 'yellow', 'olive', 'salmon',
                     'lightblue2'],
        'ggplot': ['brick', 'smurf', 'lightviolet', 'mediumgray', 'mustard', 'lime2', 'pinksalmon'],
        'polar': ['polarblue', 'polarorange', 'polargreen', 'polarpurple', 'polarred', 'polarcyan', 'polarbluelight'],
        'plotly': ['rgb(31, 119, 180)', 'rgb(255, 127, 14)', 'rgb(44, 160, 44)', 'rgb(214, 39, 40)',
                   'rgb(148, 103, 189)', 'rgb(140, 86, 75)', 'rgb(227, 119, 194)', 'rgb(127, 127, 127)',
                   'rgb(188, 189, 34)', 'rgb(23, 190, 207)'],
        'henanigans': ['henanigans_cyan2', 'henanigans_red2', 'henanigans_green2', 'henanigans_blue2',
                       'henanigans_orange2',
                       'henanigans_purple2', 'henanigans_yellow2', 'henanigans_light2', 'henanigans_cyan1',
                       'henanigans_red1',
                       'henanigans_green1', 'henanigans_blue1']
    },
    'div': {

    },
    'seq': {

    }

}

# ---------------------------------------------------------------
#  The below functions are based in colorlover by Jack Parmer
#  https://github.com/jackparmer/colorlover/
# ---------------------------------------------------------------


_scales = None
_scales_names = None


def interp(colors, N):
    def _interp(colors, N):
        try:
            return cl.interp(colors, N)
        except:
            return _interp(colors, N + 1)

    c = _interp(colors, N)
    return list(map(rgb_to_hex, cl.to_rgb(c)))


def scales(scale=None):
    """
    Displays a color scale (HTML)

    Parameters:
    -----------
            scale : str
                    Color scale name
                    If no scale name is provided then all scales are returned
                            (max number for each scale)
                    If scale='all' then all scale combinations available
                            will be returned

    Example:
            scales('accent')
            scales('all')
            scales()
    """
    if scale:
        if scale == 'all':
            display(HTML(cl.to_html(_scales)))
        else:
            display(HTML(cl.to_html(get_scales(scale))))
    else:
        s = ''
        keys = list(_scales_names.keys())
        keys.sort()
        for k in keys:
            scale = get_scales(k)
            s += '<div style="display:inline-block;padding:10px;"><div>{0}</div>{1}</div>'.format(
                k, cl.to_html(scale))
        display(HTML(s))


# Scales Dictionary
# ---------------------------------


def reset_scales():
    global _scales
    global _scales_names
    scale_cpy = cl.scales.copy()

    # Add custom scales
    for k, v in list(_custom_scales.items()):
        if v:
            for k_, v_ in list(v.items()):
                if str(len(v_)) not in scale_cpy:
                    scale_cpy[str(len(v_))] = {}
                scale_cpy[str(len(v_))][k][k_] = [
                    hex_to_rgb(normalize(_)) for _ in v_]

    # Dictionary by Type > Name > N
    _scales = {}
    for k, v in list(scale_cpy.items()):
        for k_, v_ in list(v.items()):
            if k_ not in _scales:
                _scales[k_] = {}
            for k__, v__ in list(v_.items()):
                if k__ not in _scales[k_]:
                    _scales[k_][k__] = {}
                _scales[k_][k__][k] = v__

    # Dictionary by Name > N
    _scales_names = {}
    for k, v in list(scale_cpy.items()):
        for k_, v_ in list(v.items()):
            for k__, v__ in list(v_.items()):
                k__ = k__.lower()
                if k__ not in _scales_names:
                    _scales_names[k__] = {}
                _scales_names[k__][k] = v__


def get_scales(scale=None, n=None):
    """
    Returns a color scale

    Parameters:
    -----------
            scale : str
                    Color scale name
                    If the color name is preceded by a minus (-)
                    then the scale is inversed
            n : int
                    Number of colors
                    If n < number of colors available for a given scale then
                            the minimum number will be returned
                    If n > number of colors available for a given scale then
                            the maximum number will be returned

    Example:
            get_scales('accent',8)
            get_scales('pastel1')
    """
    if scale:
        is_reverse = False
        if scale[0] == '-':
            scale = scale[1:]
            is_reverse = True
        d = copy.deepcopy(_scales_names[scale.lower()])
        keys = list(map(int, list(d.keys())))
        cs = None
        if n:
            if n in keys:
                cs = d[str(n)]
            elif n < min(keys):
                cs = d[str(min(keys))]
        if cs is None:
            cs = d[str(max(keys))]
        if is_reverse:
            cs.reverse()
        return cs
    else:
        d = {}
        for k, v in list(_scales_names.items()):
            if isinstance(v, dict):
                keys = list(map(int, list(v.keys())))
                d[k] = v[str(max(keys))]
            else:
                d[k] = v
        return d


def get_colorscale(scale):
    """
    Returns a color scale to be used for a plotly figure

    Parameters:
    -----------
            scale : str or list
                    Color scale name
                    If the color name is preceded by a minus (-)
                    then the scale is inversed.
                    Also accepts a list of colors (rgb,rgba,hex)

    Example:
            get_colorscale('accent')
            get_colorscale(['rgb(127,201,127)','rgb(190,174,212)','rgb(253,192,134)'])
    """

    if type(scale) in string_types:
        scale = get_scales(scale)
    else:
        if type(scale) != list:
            raise Exception(
                "scale needs to be either a scale name or list of colors")

    cs = [[1.0 * c / (len(scale) - 1), scale[c]] for c in range(len(scale))]
    cs.sort()
    return cs


reset_scales()



import json
import os
from . import utils


def _get_params():
    path = path = os.path.join(os.path.dirname(__file__), '../helper/params.json')
    f = open(path)
    d = json.loads(f.read())
    f.close()
    return d


def _help(figure):
    d = _get_params()
    figure = figure.lower()
    if figure in d['figures']:
        fig = {}
        fig['description'] = d['figures'][figure].get('description', '')
        fig['examples'] = d['figures'][figure].get('examples', '')
        fig['parameters'] = {figure: {'params': {}}, 'all': {'params': {}}}
        for k, v in d['parameters'].items():
            for _ in [figure, 'all']:
                if _ in v['applies']:
                    param = {}
                    if figure in v['exceptions']:
                        if v['exceptions'][figure]:
                            param['description'] = v['exceptions'][figure]
                        else:
                            break
                    else:
                        param['description'] = v['description']
                    param['type'] = v.get('type', '')
                    param['position'] = v.get('position', '')
                    if 'group' in v:
                        if v['group'] not in fig['parameters'][_]:
                            fig['parameters'][_][v['group']] = {}
                        fig['parameters'][_][v['group']][k] = param
                    else:
                        fig['parameters'][_]['params'][k] = param
        return fig
    else:
        raise Exception('Figure {0} was not found'.format(figure.upper()))


def _get_aka(figure):
    d = _get_params()
    if figure in d['figures']:
        return figure
    else:
        aka = {}
        for k, v in d['figures'].items():
            if 'aka' in v:
                for _ in d['figures'][k]['aka']:
                    aka[_] = k
        if figure in aka.keys():
            print('Did you mean: "{0}"?\n'.format(aka[figure]))
            return aka[figure]
        else:
            raise Exception('Figure {0} was not found'.format(figure.upper()))


def _printer(figure=None):
    if not figure:
        d = _get_params()
        keys = list(d['figures'].keys())
        keys.sort()
        print("Use 'blot.help(figure)' to see the list of available parameters for the given figure.")
        print("Use 'DataFrame.iplot(kind=figure)' to plot the respective figure")
        print('Figures:')
        for k in keys:
            print('\t{0}'.format(k))
    else:
        figure = _get_aka(figure)
        fig = _help(figure)
        print(figure.upper())
        if type(fig['description']) == str:
            print(fig['description'])
        else:
            for _ in fig['description']:
                print(_)
        print('\n')
        print('Parameters:\n{0}'.format('=' * 11))

        def get_tabs(val):
            return ' ' * 4 * val

        def print_params(params, tabs=0):
            _keys = list(params.keys())
            _keys.sort()
            positions = {}
            for k, v in params.items():
                if v.get('position', False):
                    positions[k] = (int(v['position']))
            for k, v in positions.items():
                if v == -1:
                    _keys.append(_keys.pop(_keys.index(k)))
                else:
                    _keys.insert(v, _keys.pop(_keys.index(k)))
            for k in _keys:
                v = params[k]
                print("{0}{1} : {2}".format(get_tabs(tabs), k, v['type']))
                if type(v['description']) == str:
                    print("{0}{1}".format(get_tabs(tabs + 1), v['description']))
                else:
                    for __ in v['description']:
                        print("{0}{1}".format(get_tabs(tabs + 1), __))

        # Params
        print_params(fig['parameters'][figure].pop('params'), 1)
        print('\n')
        print_params(fig['parameters']['all'].pop('params'), 1)
        d = utils.deep_update(fig['parameters'][figure], fig['parameters']['all'])

        for k, v in d.items():
            print('\n{0}{1}'.format(get_tabs(1), k.upper()))
            print_params(v, 2)

        if 'examples' in fig:
            print('\nEXAMPLES')
            for _ in fig['examples']:
                print('{0}>> {1}'.format(get_tabs(1), _))



"""

A productivity tool that binds pandas and plotly.
Kudos to the Plotly team!

Author: @jorgesantos

"""
from __future__ import absolute_import

from . import date_tools
from . import utils
from . import datagen
from . import helper
from . import tools
from . import colors
from . import pandastools
from . import ta
from . import themes

from .helper import _printer as help
from .plotlytools import *
from plotly.graph_objs import *
from .colors import cnames, get_colorscale
from .utils import pp
from .tools import subplots, scatter_matrix, figures, getLayout, getThemes, getTheme
from .extract import to_df
from .auth import set_config_file, get_config_file
from .quant_figure import QuantFig
from .offline import is_offline, go_offline, go_online
from .version import __version__

try:
    if get_config_file()['offline']:
        go_offline()
    else:
        go_online()
except:
    pass

# User Defined Colors

try:
    colors.cnames = utils.merge_dict(colors.cnames, auth.get_user_colors())
    colors._scales_names = utils.merge_dict(colors._scales_names, auth.get_user_scales())
    themes.THEMES = utils.merge_dict(themes.THEMES, auth.get_user_themes())
except:
    pass



import pandas as pd
import re


def _screen(self, include=True, **kwargs):
    """
    Filters a DataFrame for columns that contain the given strings.
    Parameters:
    -----------
        include : bool
            If False then it will exclude items that match
            the given filters.
            This is the same as passing a regex ^keyword
        kwargs :
            Key value pairs that indicate the column and
            value to screen for

    Example:
        df.screen(col1='string_to_match',col2=['string1','string2'])
    """
    df = self.copy()
    for k, v in list(kwargs.items()):
        v = [v] if type(v) != list else v
        if include:
            df = df[df[k].str.contains('|'.join(v), flags=re.IGNORECASE).fillna(False)]
        else:
            df = df[df[k].str.contains('|'.join(v), flags=re.IGNORECASE).fillna(False) == False]
    return df


def _swapcolumns(self):
    """
    Swaps first and second columns.
    Useful for inverting axis when plotting.

    Example:
        df.swapcolumns()

    Returns : DataFrame

    """
    return self.reindex_axis([self.columns[1], self.columns[0]], axis=1)


def bestfit(self):
    """
    Returns a series with the bestfit values.

    Example:
        Series.bestfit()

    Returns: series
        The returned series contains a parameter
        called 'formula' which includes the string representation
        of the bestfit line.
    """
    # statsmodel cannot be included on requirements.txt
    # see https://github.com/scikit-learn/scikit-learn/issues/4164
    # which shares the same issue as statsmodel
    try:
        import statsmodels.api as sm
    except:
        raise Exception("statsmodels is required: " \
                        "please run " \
                        "pip install statsmodels")

    if isinstance(self.index, pd.DatetimeIndex):
        x = pd.Series(list(range(1, len(self) + 1)), index=self.index)
    else:
        x = self.index.values

    x = sm.add_constant(x)
    model = sm.OLS(self, x)
    fit = model.fit()
    vals = fit.params.values
    best_fit = fit.fittedvalues
    # the below methods have been deprecated in Pandas
    # model=pd.ols(x=x,y=self,intercept=True)
    # best_fit=model.y_fitted
    best_fit.formula = '%.2f*x+%.2f' % (vals[1], vals[0])
    return best_fit


def normalize(self, asOf=None, multiplier=100):
    """
    Returns a normalized series or DataFrame

    Example:
        Series.normalize()

    Returns: series of DataFrame

    Parameters:
    -----------
        asOf : string
            Date format
            '2015-02-29'
        multiplier : int
            Factor by which the results will be adjusted
    """
    if not asOf:
        x0 = self.iloc[0]
    else:
        x0 = self.loc[asOf]
    return self / x0 * multiplier


def read_google(url, **kwargs):
    """
    Reads a google sheet
    """
    if url[-1] != '/':
        url += '/'
    return pd.read_csv(url + 'export?gid=0&format=csv', **kwargs)


pd.DataFrame.screen = _screen
pd.DataFrame.swapcolumns = _swapcolumns
pd.DataFrame.normalize = normalize
pd.read_google = read_google
pd.Series.normalize = normalize
pd.Series.bestfit = bestfit



import pandas as pd
import time
import copy
import numpy as np
from plotly.graph_objs import Figure, Layout, Bar, Box, Scatter, FigureWidget, Scatter3d, Histogram, Heatmap, Surface, \
    Pie
from collections import defaultdict
from .exceptions import BlotError
from .colors import normalize, get_scales, colorgen, to_rgba, get_colorscale
from .utils import check_kwargs, deep_update, kwargs_from_keyword, is_list
from . import tools
from . import offline
from . import auth

__TA_KWARGS = ['min_period', 'center', 'freq', 'how', 'rsi_upper', 'rsi_lower', 'boll_std', 'fast_period',
               'slow_period', 'signal_period', 'initial', 'af', 'open', 'high', 'low', 'close']


def iplot_to_dict(data):
    d = defaultdict(dict)
    for i in data:
        for k, v in list(i.items()):
            d[i['name']][k] = v
    return d


def dict_to_iplot(d):
    l = []
    for v in list(d.values()):
        l.append(v)
    return l


def _to_iplot(self, colors=None, colorscale=None, kind='scatter', mode='lines', interpolation='linear', symbol='dot',
              size='12', fill=None, fillcolor=None, showlegend=None,
              width=3, dash='solid', sortbars=False, keys=None, bestfit=False, bestfit_colors=None, opacity=0.6,
              asDates=False, asTimestamp=False, text=None, **kwargs):
    """
    Generates a plotly Data object

    Parameters
    ----------
        colors : list or dict
            {key:color} to specify the color for each column
            [colors] to use the colors in the defined order
        colorscale : str
            Color scale name
            Only valid if 'colors' is null
            See blot.colors.scales() for available scales
        kind : string
            Kind of chart
                scatter
                bar
        mode : string
            Plotting mode for scatter trace
                lines
                markers
                lines+markers
                lines+text
                markers+text
                lines+markers+text
        interpolation : string
            Positioning of the connecting lines
                linear
                spline
                vhv
                hvh
                vh
                hv
        symbol : string
            The symbol that is drawn on the plot for each marker
            Valid only when mode includes markers
                circle
                circle-dot
                diamond
                square
                and many more...(see plotly.validators.scatter.marker.SymbolValidator.values)
        size : string or int
            Size of marker
            Valid only if marker in mode
        fill : bool
            Filled Traces
        width : int
            Line width
        dash : string
            Drawing style of lines
                solid
                dash
                dashdot
                dot
        sortbars : boole
            Sort bars in descending order
            * Only valid when kind='bar'
        keys : list of columns
            List of columns to chart.
            Also can be used for custom sorting.
        bestfit : boolean or list
            If True then a best fit line will be generated for
            all columns.
            If list then a best fit line will be generated for
            each key on the list.
        bestfit_colors : list or dict
            {key:color} to specify the color for each column
            [colors] to use the colors in the defined order
        asDates : bool
            If true it forces truncates times from a DatetimeIndex

    """
    df = self.copy()

    if asTimestamp:
        x = [_ for _ in df.index]
    elif df.index.__class__.__name__ in ('PeriodIndex', 'DatetimeIndex'):
        if asDates:
            df.index = df.index.date
        x = df.index.format()
    elif isinstance(df.index, pd.MultiIndex):
        x = ['({0})'.format(','.join([str(__) for __ in _])) for _ in df.index.values]
    else:
        x = df.index.values
    lines = {}
    if type(df) == pd.core.series.Series:
        df = pd.DataFrame({df.name: df})

    if not keys:
        if 'bar' in kind:
            if sortbars:
                keys = list(df.sum().sort_values(ascending=False).keys())
            else:
                keys = list(df.keys())
        else:
            keys = list(df.keys())

    colors = get_colors(colors, colorscale, keys)
    dash = get_items_as_list(dash, keys, 'dash')
    symbol = get_items_as_list(symbol, keys, 'symbol')
    mode = get_items_as_list(mode, keys, 'mode')
    interpolation = get_items_as_list(interpolation, keys, 'interpolation')
    width = get_items_as_list(width, keys, 'width')

    upper, lower = kwargs.get('upper'), kwargs.get('lower')
    if upper or lower:
        bounds_color = kwargs.get('bounds_color') or 'rgba(222, 233, 252, 0.7)'
        config = {
            'x': x, 'line': {'color': bounds_color},
            'showlegend': False, 'fillcolor': bounds_color, 'mode': 'lines'
        }
        upper_data = df[upper].values if upper in df.columns else [df.max().max()] * len(x)
        lower_data = df[lower].values if lower in df.columns else [df.min().min()] * len(x)
        lines.update({
            '_upper': {'name': '_upper', 'y': upper_data, **config},
            '_lower': {'name': '_lower', 'y': lower_data, 'fill': 'tonexty', **config}
        })
        keys = [k for k in keys if k not in [upper, lower]]

    for i, key in enumerate(keys):
        lines[key] = {}
        lines[key]["x"] = x
        lines[key]["y"] = df[key].fillna('').values
        lines[key]["name"] = str(key)
        if upper or lower:
            anomaly_color = kwargs.get('anomaly_color') or 'red'
            y = df[[key]].fillna('')
            name = f'{key}_anomaly'
            y['anomaly'] = (df[key] > upper_data) | (df[key] < lower_data)
            y[key][~y['anomaly']] = np.nan
            lines[name] = {
                'name': name, 'x': x, 'y': y[key].values,
                'line': {'color': anomaly_color}, 'showlegend': False
            }

        def current_legend():
            if isinstance(showlegend, bool):
                return showlegend
            elif isinstance(fillcolor, list):
                return showlegend[i] if i < len(showlegend) else True
            elif isinstance(fillcolor, dict):
                return fillcolor.get(key, True)
            else:
                return True

        lines[key]['showlegend'] = current_legend()

        if text is not None:
            lines[key]["text"] = text
        if 'bar' in kind:
            lines[key]["marker"] = {'color': to_rgba(colors[key], opacity), 'line': {'color': colors[key], 'width': 1}}
        else:
            lines[key]["line"] = {'color': colors[key], 'width': width[key], 'dash': dash[key],
                                  'shape': interpolation[key]}
            lines[key]["mode"] = mode[key]
            if 'marker' in mode[key]:
                lines[key]["marker"] = dict(symbol=symbol[key], size=size)
            if fill:
                def current_fill():
                    if isinstance(fill, bool):
                        return 'tonexty' if kind == 'area' else 'tozeroy'
                    elif isinstance(fill, list):
                        return fill[i] if i < len(fill) else 'none'
                    elif isinstance(fill, dict):
                        return fill.get(key, 'tonexty')
                    else:
                        return fill

                def current_fillcolor():
                    default = to_rgba(colors[key], kwargs.get('opacity', 0.3))
                    if isinstance(fillcolor, str):
                        return fillcolor
                    elif isinstance(fillcolor, list):
                        return fillcolor[i] if i < len(fillcolor) else default
                    elif isinstance(fillcolor, dict):
                        return fillcolor.get(key, default)
                    else:
                        return default

                lines[key]["fill"] = current_fill()
                lines[key]["fillcolor"] = current_fillcolor()
    if 'bar' in kind:
        lines_plotly = [Bar(lines[key]).to_plotly_json() for key in lines]
    else:
        lines_plotly = [Scatter(lines[key]).to_plotly_json() for key in lines]
    for trace in lines_plotly:
        if isinstance(trace['name'], pd.Timestamp):
            trace.update(name=str(trace['name']))

    if bestfit:
        if isinstance(df.index, pd.MultiIndex):
            raise TypeError('x cannot be empty for MultiIndex dataframes')

        if type(bestfit) == list:
            keys = bestfit
        d = {}
        for key in keys:
            bestfit = df[key].bestfit()
            d[bestfit.formula] = bestfit
        bestfit_lines = pd.DataFrame(d).to_iplot(bestfit=False, colors=bestfit_colors, kind='scatter',
                                                 asTimestamp=asTimestamp)
        for line in bestfit_lines:
            line['line']['dash'] = 'dash'
            if not bestfit_colors:
                line['line']['color'] = to_rgba(line['line']['color'], .6)
        data = lines_plotly
        data.extend(bestfit_lines)
        return data

    return lines_plotly


def _iplot(
        self, kind='line', data=None, layout=None,

        x='', y='', y2='', z='', labels=None, values=None, text='', keys=None, upper='', lower='',

        title='', xtitle='', ytitle='', y2title='', ztitle='',

        theme=None, colors=None, colorscale=None, fill=None, fillcolor=None, width=None,
        dash='solid', mode='', interpolation='linear', symbol='circle', size=12, bounds_color=None, anomaly_color='red',

        barmode='', bargap=None, bargroupgap=None, bins=None, histnorm='', boxpoints=False,

        orientation='v', annotations=None, showlegend=True, vlines=None, hlines=None,

        gridcolor=None, zerolinecolor=None, bg_color=None, margin=None, dimensions=None, subplots=False, as_figure=True,
        **kwargs
):
    """
    默认返回plotly Figure对象
    Parameters:
    -----------
        kind : string
            Kind of chart
                line：折线图
                scatter：散点图
                scatter3d：3d散点图
                bar：柱状图
                histogram：直方图
                box：箱形图
                spread：折线图对比，自动增加fig1-fig2子图
                ratio：折线图对比，自动增加fig1/fig2子图
                heatmap：热力图
                surface：曲面图
                bubble：气泡图
                bubble3d：3d气泡图
                pie：扇形图

        data : Data
            Plotly Data 对象.
            如果未显式指定，则自动从DataFrame对象生成

        layout : Layout
            Plotly layout 对象
            如果未显式指定，则自动从DataFrame对象生成

        ------------------------------------

        x : string
            代表x轴取值的列名

        y : string or list
            代表y轴取值的列名

        y2 : string or list(string)
            代表第二y轴取值的列名，用法同y相同，也可以和y同时使用。

        z : string
            代表z轴取值的列名

        labels : string or list
                string：单列作为label
                list：多列的组合作为label
            Pie图中label对应的列名
            * 仅当kind='pie'时有效

        values : string
            Pie图中的取值列名
            * 仅当kind='pie'时有效

        text : string
            包含text取值的列名

        keys : list of columns
            参与绘图的df列名，默认为df.columns
            也可以用作排序，作用类似于df[keys]

        upper : string of columns
            基带上界列名

        lower : list of columns
            基带下界列名

        ------------------------------------

        title : string
            图像标题

        xtitle : string
            x轴标题

        ytitle : string
            y轴标题

        y2title : string
            第二y轴名称

        ztitle : string
            z轴标题
            仅在3d图中生效

        ------------------------------------

        theme : string
            Layout 主题
                bowaer
                solar
                pearl
                white
                …………
            通过blot.getThemes() 查看所有内置可用主题，默认主题为bowaer

        colors : dict, list or string
            {key:color} 为每一条trace单独指定颜色
            [colors] zip(colors, traces)来指定颜色

        colorscale : string
            内置的色阶名称
            若色阶名前带'-'，则颜色反转
            仅当未设置colors参数时有效
            在blot.colors.scales()查看有效色阶名称

        fill : bool, str, list or dict
                bool: False不填充。True：'tozeroy' if kind=='area' else 'tonexty'
                str: 填充方式，取值列表：( "none" | "tozeroy" | "tozerox" | "tonexty" | "tonextx" | "toself" | "tonext" )
                list：单独为每条曲线指定填充方式
                dict：同上，key of keys
            填充曲线的方式，具体参数含义参照https://plotly.com/python/reference/scatter/#scatter-fill

        fillcolor : str, list or dict
                str: 填充颜色，和color方式一样。hex(e.g. '#ff0000')、rgb/rgba(e.g. 'rgba(255,0,0,0.8)')、css color name……
                list：单独为每条曲线指定填充颜色
                dict：同上，key of keys
            曲线填充的颜色控制

        width : dict, list or int
                int : 应用于所有曲线
                list : 按照顺序应用在各条曲线
                dict: {column:value} 为每条曲线单独指定线宽
            线宽，如果不指定，则从theme读取，如果theme没有，则默认为2

        dash : dict, list or string
                string : 应用于所有曲线
                list : 按照顺序应用在各条曲线
                dict: {column:value} 为每条曲线单独指定线宽
            线型
                solid「默认」：实线  '————'
                dash：短线 '- - - - '
                dashdot：带点短线 '-•-•-•-'
                dot：点连线  '••••••'

        mode : dict, list or string
                string : 应用于所有曲线
                list : 按照顺序应用在各条曲线
                dict: {column:value} 为每条曲线单独指定模式
            散点图下绘制模式
                markers「默认」
                lines：连线。和kind='line'效果相同
                lines+markers：用线将散点连接
                lines+text：线+注解。text由text参数指定
                markers+text：点+注解
                lines+markers+text：线+点+注解

        interpolation : string
            连线方式
                linear「默认」：直线连接
                spline：曲线连接
                vhv：纵横纵的方式连接
                hvh：横纵横的方式连接
                vh：纵横的方式连接
                hv：横纵的方式连接

        symbol : string
            散点的形状，仅当mode的取值包含markers时生效
                circle「默认」：实心圆点
                circle-dot：空心圆点
                diamond：菱形
                square：正方形
                ………可参照https://plotly.com/python/reference/scatter/#scatter-marker-symbol查看更多有效值

        size : string or int
            散点大小
                string：气泡图中气泡大小对应的列名
                int：散点图中散点的大小，定值

        bounds_color : string
            基带颜色，格式同color。默认='rgba(222, 233, 252, 0.7)'

        anomaly_color : string
            异常点颜色，格式同color。默认='red'

        ------------------------------------

        barmode : string
            多维柱状图展示形式，默认group
                group
                stack
                overlay
            仅当kind='bar'时有效

        bargap : float[0,1)
            组间间隔
            仅当kind='bar'或者'histogram'时有效

        bargroupgap : float[0,1)
            组内间隔距离
            仅当kind='bar'或者'histogram'时有效

        bins : int or tuple
            int: 分桶数量
            tuple: (start, end, size)
                start : 横坐标起始截断值
                end: 横坐标结束截断值
                size: 分桶数量
            仅在kind=histogram时生效

        histnorm : string
                '' (frequency)「默认」：频数
                percent: 百分数（0-100）
                probability：频率（0-1）
                density: 密度 = frequency/bins
                probability density: 概率密度 = probability/bins
            仅在kind=histogram时生效

        boxpoints : string
            箱形图（box）中散点的展示方式。默认不展示
                False「默认」：不展示
                outliers：仅展示离群点
                all：展示所有点
                suspectedoutliers：仅展示疑似离群点

        ------------------------------------

        orientation : string
            图例方向
                v「默认」：纵向
                h：横向
            仅当kind='bar'、'histogram'、'box'时有效

        annotations : dict
            图例标注
            {x_point : text}：{横坐标取值, 标注内容}

        vlines: list
            需要绘制垂直（平行于y轴）参考线的x取值列表

        hlines: list
            需要绘制水平（平行于x轴）参考线的y取值列表

        ------------------------------------

        gridcolor : string
            网格线颜色

        zerolinecolor : string
            坐标轴颜色

        margin : dict or tuple
                Tuple: 按照（left, right, bottom, top) 给出取值
                dict: key of ('l', 'r', 'b', 't'), 例如{'l':30} 代表距离左边界30px
            图例和上下左右边界的距离

        dimensions : tuple(int,int)
            图像的尺寸(width,height)
            jupyter中默认会充满输出
            返回静态图或者plotly对象默认为(800, 500)

        subplots : bool or tuple(rows, cols)
                bool:是否以子图的形式展示，默认为False。True则按照2列的形式展示子图
                tuple(rows, cols): 子图的栅栏分布
            子图的展示控制

        as_figure : bool=True
            是否返回plotly对象


    ==========Other Kwargs===================
        Line, Scatter
            connectgaps : bool=False
               折线图中空值是否连接

        ------------------------------------

        Pie charts
            sort : bool=True
                legend中各label是否按照占比排序

            pull : float [0-1] or list(float)
                    float：每块离中心的距离，默认为0
                    list：按顺序指定每个块离中心的距离
                饼图中块距离中心点的距离

            hole : float [0-1]
                饼图中心孔的尺寸，默认为0

            textcolor : string
                饼图中显示的文字颜色，默认自适应（深色背景浅色字体，浅色背景深色背景）

            textposition : string or list（string）
                饼图中的文字显示的位置 ['inside', 'outside', 'auto', 'none']

            textinfo : string
                饼图中文字显示的信息，默认为percent
                可以为['label', 'text', 'value', 'percent']中的一个
                或者多个用'+'连接的字符串，eg. 'label+value'

        ------------------------------------

        Heatmap and Surface
            center_value : float
                指定热力图scale的中心取值
                如果不指定，将自动从(zmin,zmax)中计算

            zmin : float
                指定计算colorscale最小的取值。eg. zmin=0.5, 则z<=0.5的颜色均为colorscale的起始颜色
                会影响到整个colorscale range

            zmax : float
                 指定计算colorscale最大的取值
                 会影响到整个colorscale range

        ------------------------------------

        Subplots
            horizontal_spacing : float [0,1]
                多个水平子图之间的距离，列间距

            vertical_spacing : float [0,1]
                多个竖直子图之间的距离，行间距

            subplot_titles : bool=False
                是否显示每个子图的title（默认使用legend信息）

            shared_xaxes、shared_xaxis : bool
                子图间是否共享x轴

            shared_yaxes : bool
                子图间是否共享y轴

        ------------------------------------

        Shapes
            hpsans : tuple(y0,y1) or list(tuple)
                填充水平（平行于y轴）区域，y0,y1表示区域的起始和结束纵坐标。可以用list传入多个

            vspans : tuple(x0,x1) or list(tuple)
                填充竖直（平行于x轴）区域，y0,y1表示区域的起始和结束纵坐标。可以用list传入多个

            shapes : dict or list(dict)
                plotly中的shape属性字典.具体参照https://plotly.com/python/shapes/#vertical-and-horizontal-lines-positioned-relative-to-the-axis-data
                eg. {"type":"rect", 'x0':1, 'y0':1, 'x1':2, 'y1':3}则绘制一个长方形框

        ------------------------------------

        Axis Ranges
            xrange : [lower_bound,upper_bound]
                指定x轴取值范围

            yrange : [lower_bound,upper_bound]
                指定y轴取值范围

            zrange : [lower_bound,upper_bound]
                指定z轴取值范围

        ------------------------------------

        Range Slider
            rangeslider : bool or dict
                    bool: 默认False, 如果为True，将展示默认的range slider
                    dict: 具体参照plotly文档：https://plotly.com/python/range-slider/
                        eg.  {'bgcolor':('blue',.3),'autorange':True}
                展示range slider

        ------------------------------------

        Annotations
            fontcolor : str
                注解文字颜色

            fontsize : int
                注解文字大小

            textangle : int=0
                注解文字方向
            完成注解相关配置参考：https://plot.ly/python/reference/#layout-annotations

    """
    # Valid Kwargs
    valid_kwargs = [
        'color', 'opacity', 'column', 'columns', 'labels', 'text', 'world_readable', 'colorbar',
        'vline_color', 'vline_width', 'hline_color', 'hline_width','bgcolor'
    ]
    BUBBLE_KWARGS = ['abs_size']
    TRACE_KWARGS = ['hoverinfo', 'connectgaps']
    HEATMAP_SURFACE_KWARGS = ['center_value', 'zmin', 'zmax']
    PIE_KWARGS = ['sort', 'pull', 'hole', 'textposition', 'textinfo', 'linecolor', 'linewidth', 'textcolor']
    OHLC_KWARGS = [
        'up_color', 'down_color', 'open', 'high', 'low', 'close',
        'volume', 'name', 'decreasing', 'increasing'
    ]
    SUBPLOT_KWARGS = [
        'horizontal_spacing', 'vertical_spacing',
        'specs', 'insets', 'start_cell', 'shared_xaxes',
        'shared_yaxes', 'subplot_titles', 'shared_xaxis', 'shared_yaxis'
    ]

    EXPORT_KWARGS = ['display_image', 'scale']

    kwargs_list = [
        tools.__LAYOUT_KWARGS, BUBBLE_KWARGS, TRACE_KWARGS,
        OHLC_KWARGS, PIE_KWARGS, HEATMAP_SURFACE_KWARGS, SUBPLOT_KWARGS,
        EXPORT_KWARGS,
    ]
    [valid_kwargs.extend(_) for _ in kwargs_list]

    dict_modifiers_keys = ['line']
    dict_modifiers = {}

    for k in dict_modifiers_keys:
        dict_modifiers[k] = kwargs_from_keyword(kwargs, {}, k, False)

    for key in list(kwargs.keys()):
        if key not in valid_kwargs:
            raise Exception("Invalid keyword : '{0}'".format(key))

    # Setting default values
    if not colors:
        colors = kwargs['color'] if 'color' in kwargs else colors
    if isinstance(colors, str):
        colors = [colors]
    opacity = kwargs.get('opacity', 0.8)

    theme = theme or 'bowaer'

    theme_config = tools.getTheme(theme)

    colorscale = colorscale or theme_config.get('colorscale', 'dflt')

    if width is None and kind != 'pie':
        width = theme_config.get('linewidth', 2)

    # 常见错误参数修正
    if 'column' in kwargs:
        keys = [kwargs['column']] if isinstance(kwargs['column'], str) else kwargs['column']
    if 'columns' in kwargs:
        keys = [kwargs['columns']] if isinstance(kwargs['columns'], str) else kwargs['columns']
    kind = 'line' if kind == 'lines' else kind

    # Figure generators
    def get_marker(marker):
        marker = marker or {}
        if 'line' in dict_modifiers:
            if 'color' not in dict_modifiers['line']:
                if 'linecolor' in kwargs:
                    linecolor = kwargs.get('linecolor')
                else:
                    if 'linecolor' in tools.getTheme(theme=theme):
                        linecolor = normalize(tools.getTheme(theme=theme)['linecolor'])
                    else:
                        linecolor = tools.getLayout(theme=theme)['xaxis']['titlefont']['color']
                dict_modifiers['line']['color'] = linecolor
            dict_modifiers['line'] = tools.updateColors(dict_modifiers['line'])
            marker['line'] = deep_update(marker['line'], dict_modifiers['line'])
        return marker

    # We assume we are good citizens
    validate = True

    if not layout:
        l_kwargs = dict([(k, kwargs[k]) for k in tools.__LAYOUT_KWARGS if k in kwargs])
        for k in ['vspan', 'hspan']:
            v = l_kwargs.get(k) or l_kwargs.get(k + 's')
            if v:
                l_kwargs[k] = v

        if annotations:
            ann_kwargs = check_kwargs(kwargs, tools.__ANN_KWARGS, clean_origin=False)
            annotations = tools.get_annotations(self.copy(), annotations, kind=kind, theme=theme, **ann_kwargs)

        layout = tools.getLayout(
            kind=kind, theme=theme, xTitle=xtitle, yTitle=ytitle, zTitle=ztitle, title=title,
            barmode=barmode, bargap=bargap, bargroupgap=bargroupgap, annotations=annotations,
            gridcolor=gridcolor, dimensions=dimensions, zerolinecolor=zerolinecolor,
            margin=margin, is3d='3d' in kind, **l_kwargs
        )
    elif isinstance(layout, Layout):
        layout = layout.to_plotly_json()

    layout['shapes'] = layout.get('shapes', [])
    layout['shapes'] += tools.add_ref_line(
        vlines or [], direction='vertical',
        color=kwargs.get('vline_color', 'red'), width=kwargs.get('vline_width', 2)
    )

    bg_color = bg_color or kwargs.get('bgcolor')
    if bg_color:
        layout['legend']['bgcolor'] = bg_color
        layout['paper_bgcolor'] = bg_color
        layout['plot_bgcolor'] = bg_color


    layout['shapes'] += tools.add_ref_line(
        hlines or [], direction='horizontal',
        color=kwargs.get('hline_color', 'red'), width=kwargs.get('hline_width', 2)
    )

    if not data:
        if kind in ('scatter', 'spread', 'ratio', 'bar', 'barh', 'area', 'line'):
            df = self.copy()
            bounds_cols = list(filter(bool, [upper, lower]))
            if type(df) == pd.core.series.Series:
                df = pd.DataFrame({df.name: df})
            if x:
                df = df.set_index(x)
            if y and y2:
                _y = [y] if not isinstance(y, list) else y
                _secondary_y = [y2] if not isinstance(y2, list) else y2
                df = df[_y + _secondary_y + bounds_cols]
            elif y:
                df = df[([y] if not isinstance(y, list) else y) + bounds_cols]
            if kind == 'area':
                df = df.transpose().fillna(0).cumsum().transpose()
            mode = mode or ('lines' if kind != 'scatter' else 'markers')
            if text:
                if not isinstance(text, list):
                    text = self[text].values
            data = df.to_iplot(
                colors=colors, colorscale=colorscale, kind=kind, interpolation=interpolation,
                fill=fill, fillcolor=fillcolor, width=width, dash=dash, keys=keys,
                upper=upper, lower=lower, bounds_color=bounds_color, anomaly_color=anomaly_color,
                asDates=False, mode=mode, symbol=symbol, size=size, showlegend=showlegend,
                text=text, **kwargs
            )
            trace_kw = check_kwargs(kwargs, TRACE_KWARGS)
            for trace in data:
                trace.update(**trace_kw)

            if kind in ('spread', 'ratio'):
                if kind == 'spread':
                    trace = self.apply(lambda x: x[0] - x[1], axis=1)
                    positive = trace.apply(lambda x: x if x >= 0 else np.nan)
                    negative = trace.apply(lambda x: x if x < 0 else np.nan)
                    trace = pd.DataFrame({'positive': positive, 'negative': negative})
                    trace = trace.to_iplot(colors={'positive': 'green', 'negative': 'red'}, width=0.5)
                else:
                    trace = self.apply(lambda x: x[0] * 1.0 / x[1], axis=1).to_iplot(colors=['green'], width=1)
                for t in trace:
                    t.update({'xaxis': 'x2', 'yaxis': 'y2', 'fill': 'tozeroy',
                              'name': kind.capitalize(), 'connectgaps': False, 'showlegend': False})
                data.append(trace[0])
                if kind == 'spread':
                    data.append(trace[1])
                layout['yaxis'].update({'domain': [.3, 1]})
                layout['yaxis2'] = copy.deepcopy(layout['yaxis'])
                layout['xaxis2'] = copy.deepcopy(layout['xaxis'])
                layout['yaxis2'].update(domain=[0, .25], title=kind.capitalize())
                layout['xaxis2'].update(anchor='y2', showticklabels=False)
                layout['hovermode'] = 'x'
            if 'bar' in kind:
                if 'stack' in barmode:
                    layout['legend'].update(traceorder='normal')
                orientation = 'h' if kind == 'barh' else orientation
                for trace in data:
                    trace.update(orientation=orientation)
                    if orientation == 'h':
                        trace['x'], trace['y'] = trace['y'], trace['x']

        elif kind == 'bubble':
            mode = 'markers' if 'markers' not in mode else mode
            x = self[x].values.tolist()
            y = self[y].values.tolist()
            z = size if size else z
            rg = self[z].values
            if not kwargs.get('abs_size', False):
                if len(rg) > 1:
                    z = [int(100 * (float(_) - rg.min()) / (rg.max() - rg.min())) + 12 for _ in rg]
                else:
                    z = [12] if len(rg) else []
            else:
                z = rg
            text = kwargs['labels'] if 'labels' in kwargs else text
            labels = self[text].values.tolist() if text else ''
            clrs = colors if colors else get_scales(colorscale)
            clrs = [clrs] if not isinstance(clrs, list) else clrs
            clrs = [clrs[0]] * len(x) if len(clrs) == 1 else clrs
            marker = dict(color=clrs, size=z, symbol=symbol,
                          line=dict(width=width))
            trace = Scatter(x=x, y=y, marker=marker, mode='markers', text=labels)
            data = [trace]

        elif kind in ('box', 'histogram', 'hist'):
            if isinstance(self, pd.core.series.Series):
                df = pd.DataFrame({self.name: self})
            else:
                df = self.copy()
            data = []
            clrs = get_colors(colors, colorscale, df.columns)
            if 'hist' in kind:
                barmode = 'overlay' if barmode == '' else barmode
                layout.update(barmode=barmode)
            columns = keys if keys else df.columns
            for _ in columns:
                if kind == 'box':
                    __ = Box(y=df[_].values.tolist(), marker=dict(color=clrs[_]), name=_,
                             line=dict(width=width), boxpoints=boxpoints)
                    # 114 - Horizontal Box
                    __['orientation'] = orientation
                    if orientation == 'h':
                        __['x'], __['y'] = __['y'], __['x']

                else:
                    __ = dict(x=df[_].values.tolist(), name=_,
                              marker=dict(color=clrs[_], line=dict(width=width)),
                              orientation=orientation,
                              opacity=kwargs['opacity'] if 'opacity' in kwargs else .8,
                              histnorm=histnorm)

                    __['marker'] = get_marker(__['marker'])

                    if orientation == 'h':
                        __['y'] = __['x']
                        del __['x']
                    __ = Histogram(__)
                    if bins:
                        if type(bins) in (tuple, list):
                            try:
                                _bins = {'start': bins[0], 'end': bins[1], 'size': bins[2]}
                                if orientation == 'h':
                                    __.update(ybins=_bins, autobiny=False)
                                else:
                                    __.update(xbins=_bins, autobinx=False)
                            except:
                                print("Invalid format for bins generation")
                        else:
                            if orientation == 'h':
                                __.update(nbinsy=bins)
                            else:
                                __.update(nbinsx=bins)
                data.append(__)

        elif kind in ('heatmap', 'surface'):
            if x:
                x = self[x].values.tolist()
            else:
                if self.index.__class__.__name__ in ('PeriodIndex', 'DatetimeIndex'):
                    x = self.index.format()
                else:
                    x = self.index.values.tolist()
            y = self[y].values.tolist() if y else self.columns.values.tolist()
            z = self[z].values.tolist() if z else self.values.transpose()
            scale = get_scales('rdbu') if not colorscale else get_scales(colorscale)
            scale = [normalize(_) for _ in scale]
            colorscale = [[float(_) / (len(scale) - 1), scale[_]] for _ in range(len(scale))]
            center_value = kwargs.get('center_value', None)

            if is_list(z):
                zmin = min(z)
                zmax = max(z)
            else:
                zmin = z.min()
                zmax = z.max()

            if center_value is not None:
                if center_value <= zmin + (zmax - zmin) / 2:
                    zmin = center_value * 2 - zmax
                else:
                    zmax = center_value * 2 - zmin
            zmin = kwargs.get('zmin', zmin)
            zmax = kwargs.get('zmax', zmax)
            if kind == 'heatmap':
                data = [Heatmap(z=z, x=x, y=y, zmin=zmin, zmax=zmax, colorscale=colorscale)]
            else:
                data = [Surface(z=z, x=x, y=y, colorscale=colorscale)]

        elif kind in ('scatter3d', 'bubble3d'):
            data = []
            keys = self[text].values if text else list(range(len(self)))
            colors = get_colors(colors, colorscale, keys, asList=True)
            mode = 'markers' if 'markers' not in mode else mode
            df = self.copy()
            df['index'] = keys
            if kind == 'bubble3d':
                rg = self[size].values
                if not kwargs.get('abs_size', False):
                    size = [int(100 * (float(_) - rg.min()) / (rg.max() - rg.min())) + 12 for _ in rg]
                else:
                    size = rg
            else:
                size = [size for _ in range(len(keys))]

            _data = Scatter3d(x=df[x].values.tolist(), y=df[y].values.tolist(), z=df[z].values.tolist(), mode=mode,
                              text=keys,
                              marker=dict(color=colors, symbol=symbol, size=size, opacity=.8))
            if text:
                _data.update(text=keys)
            data.append(_data)

        elif kind == 'pie':
            if not labels:
                raise BlotError('Missing: labels')
            if not values:
                raise BlotError('Missing: values')
            labels = self[labels].values.tolist()
            values = self[values].values.tolist()
            marker = dict(colors=get_colors(colors, colorscale, labels, asList=True))
            marker.update(line=dict(color=kwargs.pop('linecolor', None), width=kwargs.pop('linewidth', width)))
            pie = dict(values=values, labels=labels, name='', marker=marker)

            kw = check_kwargs(kwargs, PIE_KWARGS)
            kw['textfont'] = {'color': kw.pop('textcolor', None)}
            pie.update(kw)
            data = []
            del layout['xaxis']
            del layout['yaxis']
            data.append(Pie(pie))
            validate = False

    filename = title or 'blot-chart'

    figure = {
        'data': data, 'layout': layout
    }

    # Check secondary axis
    if y2:
        figure = tools._set_axis(figure, y2, side='right')
        if y2title:
            figure.layout.yaxis2.title = y2title

    # Subplots
    if subplots:
        fig = tools.strip_figures(figure)
        kw = check_kwargs(kwargs, SUBPLOT_KWARGS)
        for _ in ['x', 'y']:
            if 'shared_{0}axes'.format(_) not in kw:
                kw['shared_{0}axes'.format(_)] = kw.pop('shared_{0}axis'.format(_), False)
        if 'subplot_titles' in kwargs:
            if kwargs['subplot_titles'] == True:
                kw['subplot_titles'] = [d['name'] for d in data]
            else:
                kw['subplot_titles'] = kwargs['subplot_titles']
        # 如果subplots传入参数为（rows,cols）则保留，否则交给后续自动生成形状
        subplots_shape = subplots if isinstance(subplots, tuple) and len(subplots) == 2 else None
        figure = tools.subplots(
            fig, subplots_shape,
            base_layout=layout, theme=theme, **kw
        )

    # Exports
    validate = False if 'shapes' in layout else validate

    if as_figure:
        return Figure(figure)
    else:
        return iplot(
            figure, validate=validate, filename=filename,
            dimensions=dimensions
        )


def get_colors(colors, colorscale, keys, asList=False):
    if type(colors) != dict:
        if not colors:
            if colorscale:
                colors = get_scales(colorscale, len(keys))
        clrgen = colorgen(colors, len(keys))
        if asList:
            colors = [next(clrgen) for _ in keys]
        else:
            colors = {}
            for key in keys:
                colors[key] = next(clrgen)
    return colors


def get_items_as_list(items, keys, items_names='styles'):
    """
    Returns a dict with an item per key

    Parameters:
    -----------
        items : string, list or dict
            Items (ie line styles)
        keys: list
            List of keys
        items_names : string
            Name of items
    """
    if type(items) != dict:
        if type(items) == list:
            if len(items) != len(keys):
                raise Exception('List of {0} is not the same length as keys'.format(items_names))
            else:
                items = dict(zip(keys, items))
        else:
            items = dict(zip(keys, [items] * len(keys)))
    return items


def _scatter_matrix(self, theme=None, bins=10, color='grey', size=2, asFigure=False, **iplot_kwargs):
    """
    Displays a matrix with scatter plot for each pair of
    Series in the DataFrame.
    The diagonal shows a histogram for each of the Series

    Parameters:
    -----------
        df : DataFrame
            Pandas DataFrame
        theme : string
            Theme to be used (if not the default)
        bins : int
            Number of bins to use for histogram
        color : string
            Color to be used for each scatter plot
        size : int
            Size for each marker on the scatter plot
        iplot_kwargs : key-value pairs
            Keyword arguments to pass through to `iplot`
    """
    sm = tools.scatter_matrix(self, theme=theme, bins=bins, color=color, size=size)
    if asFigure:
        return sm
    else:
        return iplot(sm, **iplot_kwargs)


def _figure(self, **kwargs):
    """
    Generates a Plotly figure for the given DataFrame

    Parameters:
    -----------
            All properties avaialbe can be seen with
            help(blot.pd.DataFrame.iplot)
    """
    kwargs['asFigure'] = True
    return self.iplot(**kwargs)


def _layout(self, **kwargs):
    """
    Generates a Plotly layout for the given DataFrame

    Parameters:
    -----------
            All properties avaialbe can be seen with
            help(blot.pd.DataFrame.iplot)
    """
    kwargs['asFigure'] = True
    return self.iplot(**kwargs)['layout']


# ONLINE
# py.iplot(filename,fileopt,sharing,world_readable)
# py.plot(filename,fileopt,auto_open,sharing,world_readable)
# py.image.ishow(figure,format,width,height,scale)
# py.image.save_as(figure,filename,format,width,height,scale)

# OFFLINE
# py_offline.iplot(figure,show_link,link_text,validate,image,filename,image_width,image_height,config)
# py_offline.plot(figure,show_link,link_text,validate,output_type,include_plotlyjs,filename,auto_open,image,image_filename,image_width,image_height,config)


def iplot(figure, validate=True, filename='', dimensions=None, **kwargs):
    """
    Plots a figure in IPython, creates an HTML or generates an Image

    figure : figure
        Plotly figure to be charted
    validate : bool
        If True then all values are validated before
        it is charted
    sharing : string
        Sets the sharing level permission
            public - anyone can see this chart
            private - only you can see this chart
            secret - only people with the link can see the chart
    filename : string
        Name to be used to save the file in the server, or as an image
    online : bool
        If True then the chart/image is rendered on the server
        even when running in offline mode.
    asImage : bool
            If True it returns an Image (png)
            In ONLINE mode:
                Image file is saved in the working directory
                    Accepts:
                        filename
                        dimensions
                        scale
                        display_image
            In OFFLINE mode:
                Image file is downloaded (downloads folder) and a
                regular plotly chart is displayed in Jupyter
                    Accepts:
                        filename
                        dimensions
    asUrl : bool
        If True the chart url/path is returned. No chart is displayed.
            If Online : the URL is returned
            If Offline : the local path is returned
    asPlot : bool
        If True the chart opens in browser
    dimensions : tuple(int,int)
        Dimensions for image
            (width,height)
    display_image : bool
        If true, then the image is displayed after it has been saved
        Requires Jupyter Notebook
        Only valid when asImage=True

    Other Kwargs
    ============
        legend : bool
            If False then the legend will not be shown
        scale : integer
            Increase the resolution of the image by `scale` amount
            Only valid when asImage=True
    """
    valid_kwargs = ['world_readable', 'legend', 'scale']

    for key in list(kwargs.keys()):
        if key not in valid_kwargs:
            raise Exception(f'Invalid keyword : {key}')

    if 'legend' in kwargs and 'layout' in figure:
        figure['layout'].update(showlegend=kwargs['legend'])

    # Filename Handling
    filename = filename or 'blot-chart'

    # Dimensions
    dimensions = dimensions or (800, 500)

    # Remove validation if shapes are present
    if 'layout' in figure:
        validate = False if 'shapes' in figure['layout'] else validate

    # iplot
    return offline.py_offline.iplot(
        figure, validate=validate, filename=filename
    )


def _ta_figure(self, **kwargs):
    """
    Generates a Plotly figure for the given DataFrame

    Parameters:
    -----------
            All properties avaialbe can be seen with
            help(blot.pd.DataFrame.iplot)
    """
    kwargs['asFigure'] = True
    return self.ta_plot(**kwargs)


def _ta_plot(self, study, periods=14, column=None, include=True, str='{name}({period})', detail=False,
             theme=None, sharing=None, filename='', asFigure=False, **iplot_kwargs):
    """
    Generates a Technical Study Chart

    Parameters:
    -----------
            study : string
                Technical Study to be charted
                    sma - 'Simple Moving Average'
                    rsi - 'R Strength Indicator'
            periods : int
                Number of periods
            column : string
                Name of the column on which the
                study will be done
            include : bool
                Indicates if the input column(s)
                should be included in the chart
            str : string
                Label factory for studies
                The following wildcards can be used:
                    {name} : Name of the column
                    {study} : Name of the study
                    {period} : Period used
                Examples:
                    'study: {study} - period: {period}'
            detail : bool
                If True the supporting data/calculations
                are included in the chart
            study_colors : string or [string]
                Colors to be used for the studies

        Study Specific Parameters
        -------------------------
        RSI
            rsi_upper : int (0,100]
                Level for the upper rsi band
                default : 70
            rsi_lower : int (0,100]
                Level for the lower rsi band
                default : 30
        CCI
            cci_upper : int
                Level for the upper cci band
                default : 100
            cci_lower : int
                Level for the lower cci band
                default : -100
        BOLL
            boll_std : int or float
                Number of standard deviations
        MACD
            fast_period : int
                Number of periods for the fast moving average
            slow_period : int
                Number of periods for the slow moving average
            signal_period : int
                Number of periods for the signal
        CORREL
            how : string
                Method for the correlation calculation
                    values
                    pct_cht
                    diff

    """

    if 'columns' in iplot_kwargs:
        column = iplot_kwargs.pop('columns')

    if 'period' in iplot_kwargs:
        periods = iplot_kwargs.pop('period')

    if 'world_readable' in iplot_kwargs:
        sharing = iplot_kwargs.pop('world_readable')

    if 'study_color' in iplot_kwargs:
        iplot_kwargs['study_colors'] = iplot_kwargs.pop('study_color')

    if sharing is None:
        sharing = auth.get_config_file()['sharing']
    if isinstance(sharing, bool):
        if sharing:
            sharing = 'public'
        else:
            sharing = 'private'
    iplot_kwargs['sharing'] = sharing
    if theme is None:
        theme = iplot_kwargs.pop('study_theme', auth.get_config_file()['theme'])

    if not filename:
        if 'title' in iplot_kwargs:
            filename = iplot_kwargs['title']
        else:
            filename = 'Plotly Playground {0}'.format(time.strftime("%Y-%m-%d %H:%M:%S"))

    def get_subplots(figures):
        shape = (len(figures), 1)
        layout = tools.get_base_layout(figures)
        subplots = tools.subplots(figures, shape=shape, shared_xaxes=True, base_layout=layout)
        if len(figures) == 2:
            subplots['layout']['yaxis']['domain'] = [.27, 1.0]
            subplots['layout']['yaxis2']['domain'] = [0, .25]
        return subplots

    def get_study(df, func, iplot_kwargs, iplot_study_kwargs, str=None, include=False, column=None, inset=False):
        df = df.copy()
        if inset:
            if not column:
                if isinstance(df, pd.DataFrame):
                    column = df.keys().tolist()
                else:
                    df = pd.DataFrame(df)
                    column = df.keys().tolist()
        if 'legend' in iplot_kwargs:
            iplot_study_kwargs['legend'] = iplot_kwargs['legend']
        fig_0 = df.figure(**iplot_kwargs)
        df_ta = func(df, column=column, include=False, str=str, **study_kwargs)
        kind = iplot_kwargs['kind'] if 'kind' in iplot_kwargs else ''
        iplot_study_kwargs['kind'] = 'scatter'
        iplot_study_kwargs['colors'] = iplot_study_kwargs.get('colors',
                                                              ['blue', 'green', 'red'] if study == 'dmi' else 'blue')
        fig_1 = df_ta.figure(theme=theme, **iplot_study_kwargs)
        if kind in ['candle', 'ohlc']:
            for i in fig_1['data']:
                i['x'] = [pd.Timestamp(_) for _ in i['x']]
        if inset:
            figure = tools.merge_figures([fig_0, fig_1]) if include else fig_1
        else:
            figure = get_subplots([fig_0, fig_1]) if include else fig_1
        return figure

    study_kwargs = {}
    iplot_study_kwargs = {}

    study_kwargs = check_kwargs(iplot_kwargs, __TA_KWARGS, {}, clean_origin=True)
    iplot_study_kwargs = kwargs_from_keyword(iplot_kwargs, {}, 'study')

    study_kwargs.update({'periods': periods})

    ta_func = eval('ta.{0}'.format(study))

    inset = study in ('sma', 'boll', 'ema', 'atr', 'ptps')
    figure = get_study(self, ta_func, iplot_kwargs, iplot_study_kwargs, include=include,
                       column=column, str=str, inset=inset)

    ## Add Bands
    if study in ('rsi', 'cci'):
        bands = {'rsi': (30, 70),
                 'cci': (-100, 100)}
        _upper = study_kwargs.get('{0}_upper'.format(study), bands[study][0])
        _lower = study_kwargs.get('{0}_lower'.format(study), bands[study][1])
        yref = 'y2' if include else 'y1'
        shapes = [tools.get_shape(y=i, yref=yref, color=j, dash='dash') for (i, j) in
                  [(_lower, 'green'), (_upper, 'red')]]
        figure['layout']['shapes'] = shapes

    # if study=='rsi':
    # 	rsi_upper=study_kwargs.get('rsi_upper',70)
    # 	rsi_lower=study_kwargs.get('rsi_lower',30)
    # 	yref='y2' if include else 'y1'
    # 	shapes=[tools.get_shape(y=i,yref=yref,color=j,dash='dash') for (i,j) in [(rsi_lower,'green'),(rsi_upper,'red')]]
    # 	figure['layout']['shapes']=shapes

    # if study=='cci':
    # 	cci_upper=study_kwargs.get('cci_upper',100)
    # 	cci_lower=study_kwargs.get('cci_lower',-100)
    # 	yref='y2' if include else 'y1'
    # 	shapes=[tools.get_shape(y=i,yref=yref,color=j,dash='dash') for (i,j) in [(cci_lower,'green'),(cci_upper,'red')]]
    # 	figure['layout']['shapes']=shapes

    ## Exports

    if asFigure:
        return figure
    else:
        return iplot(figure, sharing=sharing, filename=filename)


def _fig_iplot(self, validate=True, sharing=None, filename='',
               online=None, asImage=False, asUrl=False, asPlot=False,
               dimensions=None, display_image=True, **kwargs):
    """
    Plots a figure in IPython

    figure : figure
        Plotly figure to be charted
    validate : bool
        If True then all values are validated before
        it is charted
    sharing : string
        Sets the sharing level permission
            public - anyone can see this chart
            private - only you can see this chart
            secret - only people with the link can see the chart
    filename : string
        Name to be used to save the file in the server, or as an image
    online : bool
        If True then the chart is rendered on the server
        even when running in offline mode.
    asImage : bool
            If True it returns an Image (png)
            In ONLINE mode:
                Image file is saved in the working directory
                    Accepts:
                        filename
                        dimensions
                        scale
                        display_image
            In OFFLINE mode:
                Image file is downloaded (downloads folder) and a
                regular plotly chart is displayed in Jupyter
                    Accepts:
                        filename
                        dimensions
    asUrl : bool
        If True the chart url/path is returned. No chart is displayed.
            If Online : the URL is returned
            If Offline : the local path is returned
    asPlot : bool
        If True the chart opens in browser
    dimensions : tuple(int,int)
        Dimensions for image
            (width,height)
    display_image : bool
        If true, then the image is displayed after it has been saved
        Requires Jupyter Notebook
        onlh valide when asImage=True

    Other Kwargs
    ============

        legend : bool
            If False then the legend will not be shown
    """
    return iplot(self, validate=validate, sharing=sharing, filename=filename,
                 online=online, asImage=asImage, asUrl=asUrl, asPlot=asPlot,
                 dimensions=dimensions, display_image=display_image, **kwargs)


def figure_add(self: Figure, other: Figure):
    return self.add_traces(other.data)


pd.DataFrame.to_iplot = _to_iplot
pd.DataFrame.scatter_matrix = _scatter_matrix
pd.DataFrame.figure = _figure
pd.DataFrame.layout = _layout
pd.DataFrame.ta_plot = _ta_plot
pd.DataFrame.blot = _iplot
pd.DataFrame.ta_figure = _ta_figure
pd.Series.ta_figure = _ta_figure
pd.Series.ta_plot = _ta_plot
pd.Series.figure = _figure
pd.Series.to_iplot = _to_iplot
pd.Series.blot = _iplot

Figure.iplot = _fig_iplot
Figure.__add__ = figure_add
Figure.save = Figure.write_image
